import pytest
import dagster
from dagster import config
from dagster.core import types
from dagster.core.definitions import (
    OutputDefinition,
    InputDefinition,
    ArgumentDefinition,
)
from dagster.core.errors import DagsterInvalidDefinitionError
from dagster.core.decorators import solid, source, materialization, with_context
from dagster.core.execution import (
    output_single_solid,
    execute_single_solid,
    ExecutionContext,
)

# This file tests a lot of parameter name stuff
# So these warnings are spurious
# unused variables, unused arguments
# pylint: disable=W0612, W0613


def create_test_context():
    return ExecutionContext()


def create_empty_test_env():
    return config.Environment(sources={})


def test_solid():
    @solid()
    def hello_world():
        return {'foo': 'bar'}

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert result.success

    assert result.transformed_value['foo'] == 'bar'


def test_solid_with_name():
    @solid(name="foobar")
    def hello_world():
        return {'foo': 'bar'}

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert result.success

    assert result.transformed_value['foo'] == 'bar'


def test_solid_with_context():
    @solid(name="foobar")
    @with_context
    def hello_world(_context):
        return {'foo': 'bar'}

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=create_empty_test_env(),
    )

    assert result.success

    assert result.transformed_value['foo'] == 'bar'


def test_solid_with_input():
    @source(name="TEST", argument_def_dict={'foo': ArgumentDefinition(types.String)})
    def test_source(foo):
        return {'foo': foo}

    @solid(inputs=[InputDefinition(name="foo_to_foo", sources=[test_source])])
    def hello_world(foo_to_foo):
        return foo_to_foo

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=config.Environment(
            sources={
                'hello_world': {
                    'foo_to_foo': config.Source(name='TEST', args={'foo': 'bar'})
                }
            }
        ),
    )

    assert result.success

    assert result.transformed_value['foo'] == 'bar'


def test_sources():
    @source(name="WITH_CONTEXT", argument_def_dict={'foo': ArgumentDefinition(types.String)})
    @with_context
    def context_source(_context, foo):
        return {'foo': foo}

    @source(name="NO_CONTEXT", argument_def_dict={'foo': ArgumentDefinition(types.String)})
    def no_context_source(foo):
        return {'foo': foo}

    @solid(inputs=[InputDefinition(name="i", sources=[context_source, no_context_source])])
    def hello_world(i):
        return i

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=config.Environment(
            sources={'hello_world': {
                'i': config.Source(name='NO_CONTEXT', args={'foo': 'bar'})
            }}
        ),
    )

    assert result.success

    assert result.transformed_value['foo'] == 'bar'

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=config.Environment(
            sources={'hello_world': {
                'i': config.Source(name='NO_CONTEXT', args={'foo': 'bar'})
            }}
        ),
    )

    assert result.success

    assert result.transformed_value['foo'] == 'bar'


def test_materializations():
    test_output = {}

    @materialization(name="CONTEXT", argument_def_dict={'foo': ArgumentDefinition(types.String)})
    @with_context
    def materialization_with_context(_context, data, foo):
        test_output['test'] = data

    @materialization(name="NO_CONTEXT", argument_def_dict={'foo': ArgumentDefinition(types.String)})
    def materialization_no_context(data, foo):
        test_output['test'] = data

    @solid(
        output=OutputDefinition(
            materializations=[materialization_with_context, materialization_no_context]
        )
    )
    def hello():
        return {'foo': 'bar'}

    output_single_solid(
        create_test_context(),
        hello,
        environment=config.Environment(sources={}),
        name='CONTEXT',
        arg_dict={'foo': 'bar'}
    )

    assert test_output['test'] == {'foo': 'bar'}

    test_output = {}

    output_single_solid(
        create_test_context(),
        hello,
        environment=config.Environment(sources={}),
        name='NO_CONTEXT',
        arg_dict={'foo': 'bar'}
    )

    assert test_output['test'] == {'foo': 'bar'}


def test_solid_definition_errors():
    @source(name="test_source")
    def test_source():
        pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @solid(
            inputs=[InputDefinition(name="foo", sources=[test_source])], output=OutputDefinition()
        )
        @with_context
        def vargs(_context, foo, *args):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @solid(
            inputs=[InputDefinition(name="foo", sources=[test_source])], output=OutputDefinition()
        )
        def wrong_name(bar):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @solid(
            inputs=[
                InputDefinition(name="foo", sources=[test_source]),
                InputDefinition(name="bar", sources=[test_source])
            ],
            output=OutputDefinition()
        )
        def wrong_name_2(foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @solid(
            inputs=[InputDefinition(name="foo", sources=[test_source])], output=OutputDefinition()
        )
        @with_context
        def no_context(foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @solid(
            inputs=[InputDefinition(name="foo", sources=[test_source])], output=OutputDefinition()
        )
        def yes_context(_context, foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @solid(
            inputs=[InputDefinition(name="foo", sources=[test_source])], output=OutputDefinition()
        )
        def extras(foo, bar):
            pass

    @solid(
        inputs=[
            InputDefinition(name="foo", sources=[test_source]),
            InputDefinition(name="bar", sources=[test_source])
        ],
        output=OutputDefinition()
    )
    def valid_kwargs(**kwargs):
        pass

    @solid(
        inputs=[
            InputDefinition(name="foo", sources=[test_source]),
            InputDefinition(name="bar", sources=[test_source])
        ],
        output=OutputDefinition()
    )
    def valid(foo, bar):
        pass

    @solid(
        inputs=[
            InputDefinition(name="foo", sources=[test_source]),
            InputDefinition(name="bar", sources=[test_source])
        ],
        output=OutputDefinition()
    )
    @with_context
    def valid_rontext(context, foo, bar):
        pass

    @solid(
        inputs=[
            InputDefinition(name="foo", sources=[test_source]),
            InputDefinition(name="bar", sources=[test_source])
        ],
        output=OutputDefinition()
    )
    @with_context
    def valid_context_2(_context, foo, bar):
        pass


def test_source_definition_errors():
    with pytest.raises(DagsterInvalidDefinitionError):

        @source(argument_def_dict={'foo': ArgumentDefinition(types.String)})
        @with_context
        def vargs(context, foo, *args):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @source(argument_def_dict={'foo': ArgumentDefinition(types.String)})
        def wrong_name(bar):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @source(
            argument_def_dict={
                'foo': ArgumentDefinition(types.String),
                'bar': ArgumentDefinition(types.String)
            }
        )
        def wrong_name_2(foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @source(argument_def_dict={'foo': ArgumentDefinition(types.String)})
        @with_context
        def no_context(foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @source(argument_def_dict={'foo': ArgumentDefinition(types.String)})
        def yes_context(context, foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @source(argument_def_dict={'foo': ArgumentDefinition(types.String)})
        def extras(foo, bar):
            pass

    @source(argument_def_dict={'foo': ArgumentDefinition(types.String)})
    def valid_kwargs(**kwargs):
        pass

    @source(
        argument_def_dict={
            'foo': ArgumentDefinition(types.String),
            'bar': ArgumentDefinition(types.String)
        }
    )
    def valid(foo, bar):
        pass

    @source(
        argument_def_dict={
            'foo': ArgumentDefinition(types.String),
            'bar': ArgumentDefinition(types.String)
        }
    )
    @with_context
    def valid_rontext(context, foo, bar):
        pass

    @source(
        argument_def_dict={
            'foo': ArgumentDefinition(types.String),
            'bar': ArgumentDefinition(types.String)
        }
    )
    @with_context
    def valid_context_2(_context, foo, bar):
        pass


def test_materialization_definition_errors():
    with pytest.raises(DagsterInvalidDefinitionError):

        @materialization(argument_def_dict={'foo': ArgumentDefinition(types.String)})
        @with_context
        def no_data(context, foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @materialization(argument_def_dict={'foo': ArgumentDefinition(types.String)})
        @with_context
        def vargs(context, data, foo, *args):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @materialization(argument_def_dict={'foo': ArgumentDefinition(types.String)})
        def wrong_name(data, bar):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @materialization(
            argument_def_dict={
                'foo': ArgumentDefinition(types.String),
                'bar': ArgumentDefinition(types.String)
            }
        )
        def wrong_name_2(data, foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @materialization(argument_def_dict={'foo': ArgumentDefinition(types.String)})
        @with_context
        def no_context(data, foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @materialization(argument_def_dict={'foo': ArgumentDefinition(types.String)})
        def yes_context(context, data, foo):
            pass

    with pytest.raises(DagsterInvalidDefinitionError):

        @materialization(argument_def_dict={'foo': ArgumentDefinition(types.String)})
        def extras(data, foo, bar):
            pass

    @materialization(argument_def_dict={'foo': ArgumentDefinition(types.String)})
    def valid_kwargs(data, **kwargs):
        pass

    @materialization(
        argument_def_dict={
            'foo': ArgumentDefinition(types.String),
            'bar': ArgumentDefinition(types.String)
        }
    )
    def valid(data, foo, bar):
        pass

    @materialization(
        argument_def_dict={
            'foo': ArgumentDefinition(types.String),
            'bar': ArgumentDefinition(types.String)
        }
    )
    @with_context
    def valid_rontext(context, data, foo, bar):
        pass

    @materialization(
        argument_def_dict={
            'foo': ArgumentDefinition(types.String),
            'bar': ArgumentDefinition(types.String)
        }
    )
    @with_context
    def valid_context_2(_context, _data, foo, bar):
        pass


def test_wrong_argument_to_pipeline():
    def non_solid_func():
        pass

    with pytest.raises(
        DagsterInvalidDefinitionError, match='You have passed a lambda or function non_solid_func'
    ):
        dagster.PipelineDefinition(solids=[non_solid_func])

    with pytest.raises(
        DagsterInvalidDefinitionError, match='You have passed a lambda or function <lambda>'
    ):
        dagster.PipelineDefinition(solids=[lambda x: x])


def test_descriptions():
    @solid(description='foo')
    def solid_desc():
        pass

    assert solid_desc.description == 'foo'

    @source(description='bar')
    def source_desc():
        pass

    assert source_desc.description == 'bar'

    @materialization(description='baaz')
    def materialization_desc():
        pass

    assert materialization_desc.description == 'baaz'



import pandas as pd

import dagster
from dagster import config
from dagster.core import types
from dagster.core.definitions import (
    InputDefinition,
    MaterializationDefinition,
    OutputDefinition,
    SolidDefinition,
    SourceDefinition,
    ArgumentDefinition,
)
from dagster.core.execution import (ExecutionContext, execute_single_solid)
from dagster.utils.test import script_relative_path


def create_test_context():
    return ExecutionContext()


def create_hello_world_solid_no_api():
    def hello_world_transform_fn(_context, args):
        num_df = args['num_df']
        num_df['sum'] = num_df['num1'] + num_df['num2']
        return num_df

    csv_input = InputDefinition(
        name='num_df',
        sources=[
            SourceDefinition(
                source_type='CSV',
                argument_def_dict={'path': ArgumentDefinition(types.Path)},
                source_fn=lambda context, arg_dict: pd.read_csv(arg_dict['path']),
            ),
        ],
    )

    csv_materialization = MaterializationDefinition(
        name='CSV',
        materialization_fn=lambda df, arg_dict: df.to_csv(arg_dict['path'], index=False),
        argument_def_dict={'path': ArgumentDefinition(types.Path)}
    )

    hello_world = SolidDefinition(
        name='hello_world',
        inputs=[csv_input],
        transform_fn=hello_world_transform_fn,
        output=OutputDefinition(materializations=[csv_materialization])
    )
    return hello_world


def test_hello_world_no_library_support():
    hello_world = create_hello_world_solid_no_api()

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=config.Environment(
            sources={
                'hello_world': {
                    'num_df': config.Source('CSV', {'path': script_relative_path('num.csv')})
                }
            }
        ),
    )

    assert result.success

    assert result.transformed_value.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
    }


def create_dataframe_input(name):
    return InputDefinition(
        name=name,
        sources=[
            SourceDefinition(
                source_type='CSV',
                argument_def_dict={'path': ArgumentDefinition(types.Path)},
                source_fn=lambda context, arg_dict: pd.read_csv(arg_dict['path']),
            ),
        ],
    )


def create_dataframe_dependency(name, depends_on):
    return InputDefinition(
        name=name,
        sources=[
            SourceDefinition(
                source_type='CSV',
                argument_def_dict={'path': ArgumentDefinition(types.Path)},
                source_fn=lambda context, arg_dict: pd.read_csv(arg_dict['path']),
            ),
        ],
        depends_on=depends_on,
    )


def create_dataframe_output():
    def mat_fn(_context, arg_dict, df):
        df.to_csv(arg_dict['path'], index=False)

    return OutputDefinition(
        materializations=[
            MaterializationDefinition(
                name='CSV',
                materialization_fn=mat_fn,
                argument_def_dict={'path': ArgumentDefinition(types.Path)},
            ),
        ],
    )


def create_hello_world_solid_composed_api():
    def transform_fn(_context, args):
        num_df = args['num_df']
        num_df['sum'] = num_df['num1'] + num_df['num2']
        return num_df

    hello_world = SolidDefinition(
        name='hello_world',
        inputs=[create_dataframe_input(name='num_df')],
        transform_fn=transform_fn,
        output=create_dataframe_output(),
    )
    return hello_world


def test_hello_world_composed():
    hello_world = create_hello_world_solid_composed_api()

    result = execute_single_solid(
        create_test_context(),
        hello_world,
        environment=config.Environment(
            sources={
                'hello_world': {
                    'num_df': config.Source('CSV', {'path': script_relative_path('num.csv')})
                }
            }
        ),
    )

    assert result.success

    assert result.transformed_value.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
    }


def test_pipeline():
    def solid_one_transform(_context, args):
        num_df = args['num_df']
        num_df['sum'] = num_df['num1'] + num_df['num2']
        return num_df

    solid_one = SolidDefinition(
        name='solid_one',
        inputs=[create_dataframe_input(name='num_df')],
        transform_fn=solid_one_transform,
        output=create_dataframe_output(),
    )

    def solid_two_transform(_context, args):
        sum_df = args['sum_df']
        sum_df['sum_sq'] = sum_df['sum'] * sum_df['sum']
        return sum_df

    solid_two = SolidDefinition(
        name='solid_two',
        inputs=[create_dataframe_dependency(name='sum_df', depends_on=solid_one)],
        transform_fn=solid_two_transform,
        output=create_dataframe_output(),
    )

    pipeline = dagster.PipelineDefinition(solids=[solid_one, solid_two])

    environment = config.Environment(
        sources={
            'solid_one': {
                'num_df': config.Source('CSV', {'path': script_relative_path('num.csv')})
            }
        }
    )

    execute_pipeline_result = dagster.execute_pipeline(
        pipeline,
        environment=environment,
    )

    assert execute_pipeline_result.result_named('solid_two').transformed_value.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
        'sum_sq': [9, 49],
    }

    sum_sq_path_args = {'path': '/tmp/sum_sq.csv'}
    environment_two = config.Environment(
        sources={
            'solid_one': {
                'num_df': config.Source('CSV', {'path': script_relative_path('num.csv')})
            }
        },
        materializations=[
            config.Materialization(
                solid='solid_two',
                name='CSV',
                args=sum_sq_path_args,
            ),
        ]
    )

    dagster.execute_pipeline(pipeline, environment=environment_two)

    sum_sq_df = pd.read_csv('/tmp/sum_sq.csv')

    assert sum_sq_df.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
        'sum_sq': [9, 49],
    }



import pytest

from dagster import check

from dagster.utils.compatability import create_custom_source_input

from dagster.core.execution import (
    _read_source,
    _execute_core_transform,
    _execute_materialization,
    ExecutionContext,
    MaterializationDefinition,
)

from dagster.core.errors import DagsterUserCodeExecutionError


def create_test_context():
    return ExecutionContext()


def test_basic_source_runtime_error_handling():
    def source_fn_inst(_context, _arg_dict):
        raise Exception('a user error')

    erroring_input = create_custom_source_input(
        name='some_input', source_fn=source_fn_inst, argument_def_dict={}
    )

    with pytest.raises(DagsterUserCodeExecutionError):
        _read_source(create_test_context(), erroring_input.sources[0], {})


def test_basic_core_transform_error_handling():
    def transform_fn(an_input):
        check.str_param(an_input, 'an_input')
        raise Exception('exception during core transform')

    with pytest.raises(DagsterUserCodeExecutionError):
        _execute_core_transform(
            create_test_context(),
            transform_fn,
            {'an_input': 'value'},
        )


def test_basic_materialization_runtime_error_handling():
    def materialization_fn_inst(_data, arg_dict):
        assert arg_dict == {}
        raise Exception('error during output')

    materialization_def = MaterializationDefinition(
        name='CUSTOM', materialization_fn=materialization_fn_inst, argument_def_dict={}
    )

    with pytest.raises(DagsterUserCodeExecutionError):
        _execute_materialization(
            create_test_context(), materialization_def, arg_dict={}, value='whatever'
        )



import dagster
from dagster.core.definitions import InputDefinition
from dagster.core.decorators import solid


@solid()
def step_one_no_external_source():
    return {'foo': 'bar'}


@solid(inputs=[InputDefinition(name='foo_bar', depends_on=step_one_no_external_source)])
def step_two(foo_bar):
    foo_bar['foo'] = foo_bar['foo'] + foo_bar['foo']
    return foo_bar


def test_pipeline_meta():
    pipeline = dagster.PipelineDefinition(solids=[step_one_no_external_source, step_two])
    assert list(pipeline.external_inputs) == []



from six import (string_types, integer_types)

from dagster import check

import dagster


class DagsterType:
    def __init__(self, name, description=None):
        self.name = check.str_param(name, 'name')
        self.description = check.opt_str_param(description, 'description')

    def is_python_valid_value(self, _value):
        check.not_implemented('Must implement in subclass')


class DagsterScalarType(DagsterType):
    pass


class _DagsterAnyType(DagsterType):
    def __init__(self):
        super().__init__(
            name='Any', description='The type that allows any value, including no value.'
        )

    def is_python_valid_value(self, _value):
        return True


def _nullable_isinstance(value, typez):
    return value is None or isinstance(value, typez)


class PythonObjectType(DagsterType):
    def __init__(
        self,
        name,
        python_type,
        default_sources=None,
        default_materializations=None,
        description=None,
    ):
        super().__init__(name, description)
        self.python_type = check.type_param(python_type, 'python_type')
        self.default_sources = check.opt_list_param(
            default_sources, 'default_sources', of_type=dagster.SourceDefinition
        )
        self.default_materializations = check.opt_list_param(
            default_materializations,
            'default_materializations',
            of_type=dagster.MaterializationDefinition
        )

    def is_python_valid_value(self, value):
        return _nullable_isinstance(value, self.python_type)


class _DagsterStringType(DagsterScalarType):
    def is_python_valid_value(self, value):
        return _nullable_isinstance(value, string_types)


class _DagsterIntType(DagsterScalarType):
    def __init__(self):
        super().__init__('Int', description='An integer.')

    def is_python_valid_value(self, value):
        return _nullable_isinstance(value, integer_types)


class _DagsterBoolType(DagsterScalarType):
    def __init__(self):
        super().__init__('Bool', description='A boolean.')

    def is_python_valid_value(self, value):
        return _nullable_isinstance(value, bool)


String = _DagsterStringType(name='String', description='A string.')
Path = _DagsterStringType(
    name='Path',
    description='''
A string the represents a path. It is very useful for some tooling
to know that a string indeed represents a file path. That way they
can, for example, make the paths relative to a different location
for a particular execution environment.
''',
)
Int = _DagsterIntType()
Bool = _DagsterBoolType()
Any = _DagsterAnyType()



import pandas as pd

import dagster
from dagster import config
from dagster.core.execution import (execute_single_solid, output_single_solid)
from dagster.core.decorators import solid
from dagster.utils.test import (script_relative_path, get_temp_file_name)

import dagster.pandas_kernel as dagster_pd


def create_num_csv_environment():
    return config.Environment(
        sources={
            'hello_world': {
                'num_csv': config.Source('CSV', {'path': script_relative_path('num.csv')})
            }
        }
    )


def test_hello_world_with_dataframe_fns():
    hello_world = create_definition_based_solid()
    run_hello_world(hello_world)


def run_hello_world(hello_world):
    result = execute_single_solid(
        dagster.ExecutionContext(),
        hello_world,
        environment=create_num_csv_environment(),
    )

    assert result.success

    assert result.transformed_value.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
    }

    with get_temp_file_name() as temp_file_name:
        output_result = output_single_solid(
            dagster.ExecutionContext(),
            hello_world,
            environment=create_num_csv_environment(),
            name='CSV',
            arg_dict={'path': temp_file_name},
        )

        assert output_result.success

        assert pd.read_csv(temp_file_name).to_dict('list') == {
            'num1': [1, 3],
            'num2': [2, 4],
            'sum': [3, 7],
        }


def create_definition_based_solid():
    table_input = dagster.InputDefinition('num_csv', dagster_pd.DataFrame)

    def transform_fn(_context, args):
        num_csv = args['num_csv']
        num_csv['sum'] = num_csv['num1'] + num_csv['num2']
        return num_csv

    # supports CSV and PARQUET by default
    hello_world = dagster.SolidDefinition(
        name='hello_world',
        inputs=[table_input],
        transform_fn=transform_fn,
        output=dagster.OutputDefinition(dagster_pd.DataFrame)
    )
    return hello_world


def create_decorator_based_solid():
    @solid(
        inputs=[dagster.InputDefinition('num_csv', dagster_pd.DataFrame)],
        output=dagster.OutputDefinition(dagster_pd.DataFrame),
    )
    def hello_world(num_csv):
        num_csv['sum'] = num_csv['num1'] + num_csv['num2']
        return num_csv

    return hello_world


def test_hello_world_decorator_style():
    hello_world = create_decorator_based_solid()
    run_hello_world(hello_world)
    result = execute_single_solid(
        dagster.ExecutionContext(),
        hello_world,
        environment=create_num_csv_environment(),
    )

    assert result.success

    assert result.transformed_value.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
    }



import sqlalchemy
import dagster.sqlalchemy_kernel as dagster_sa


# disable warnings about malformed inheritance
# pylint: disable=W0221, W0223
class MockEngine(sqlalchemy.engine.Engine):
    def __init__(self):
        super().__init__(None, None, None)

    def connect(self):
        raise Exception('should not call')

    def raw_connection(self):
        raise Exception('should not call')


def test_mock():
    sa_resource = dagster_sa.SqlAlchemyResource(engine=MockEngine(), mock_sql=True)
    context = dagster_sa.common.create_sql_alchemy_context_from_sa_resource(sa_resource)
    dagster_sa.common.execute_sql_text_on_context(context, 'NOPE')


import os
import dagster
import dagster.sqlalchemy_kernel as dagster_sa

from dagster import InputDefinition


def _get_sql_script_path(name):
    return os.path.join(os.path.dirname(__file__), 'sql_files', f'{name}.sql')


def _get_project_solid(name, inputs=None):
    return dagster_sa.sql_file_solid(_get_sql_script_path(name), inputs=inputs)


def define_full_pipeline():
    create_all_tables_solids = _get_project_solid('create_all_tables')

    populate_num_table_solid = _get_project_solid(
        'populate_num_table',
        inputs=[
            InputDefinition(create_all_tables_solids.name, depends_on=create_all_tables_solids)
        ],
    )

    insert_deps = [
        InputDefinition(populate_num_table_solid.name, depends_on=populate_num_table_solid)
    ]

    insert_into_sum_table_solid, insert_into_sum_sq_table_solid = get_insert_solids(
        insert_deps=insert_deps
    )

    return dagster.PipelineDefinition(
        name='full_pipeline',
        description='Runs entire pipeline, both setup and running the transform',
        solids=[
            create_all_tables_solids,
            populate_num_table_solid,
            insert_into_sum_table_solid,
            insert_into_sum_sq_table_solid,
        ],
    )


def get_insert_solids(insert_deps):
    insert_into_sum_table_solid = _get_project_solid(
        'insert_into_sum_table',
        inputs=insert_deps,
    )

    insert_into_sum_sq_table_solid = _get_project_solid(
        'insert_into_sum_sq_table',
        inputs=[
            InputDefinition(
                insert_into_sum_table_solid.name, depends_on=insert_into_sum_table_solid
            )
        ],
    )
    return insert_into_sum_table_solid, insert_into_sum_sq_table_solid


def define_truncate_pipeline():
    truncate_solid = _get_project_solid('truncate_all_derived_tables')
    return dagster.PipelineDefinition(
        name='truncate_all_derived_tables',
        description=
        'Truncates all tables that are populated by the pipeline. Preserves source tables',
        solids=[truncate_solid]
    )


def define_rerun_pipeline():
    insert_into_sum_table_solid, insert_into_sum_sq_table_solid = get_insert_solids(
        insert_deps=None
    )

    return dagster.PipelineDefinition(
        name='rerun_pipeline',
        description=
        'Rerun the pipeline, populating the the derived tables. Assumes pipeline is setup',
        solids=[
            insert_into_sum_table_solid,
            insert_into_sum_sq_table_solid,
        ],
    )


def define_setup_pipeline():
    create_all_tables_solids = _get_project_solid('create_all_tables')

    populate_num_table_solid = _get_project_solid(
        'populate_num_table',
        inputs=[
            InputDefinition(create_all_tables_solids.name, depends_on=create_all_tables_solids)
        ]
    )

    return dagster.PipelineDefinition(
        name='setup_pipeline',
        description='Creates all tables and then populates source table',
        solids=[
            create_all_tables_solids,
            populate_num_table_solid,
        ],
    )



from dagster.core.types import (DagsterType, PythonObjectType)


def test_desc():
    type_foo = DagsterType(name='Foo', description='A foo')
    assert type_foo.name == 'Foo'
    assert type_foo.description == 'A foo'


def test_python_object_type():
    class Bar:
        pass

    type_bar = PythonObjectType('Bar', Bar, description='A bar.')

    assert type_bar.name == 'Bar'
    assert type_bar.description == 'A bar.'
    assert type_bar.is_python_valid_value(Bar())
    assert type_bar.is_python_valid_value(None)  # allow nulls
    assert not type_bar.is_python_valid_value('not_a_bar')



from dagster import check
from dagster import config

import dagster
import dagster.core
from dagster.core.definitions import (
    SolidDefinition,
    InputDefinition,
)
from dagster.core.execution import (
    DagsterExecutionFailureReason,
    execute_pipeline,
    ExecutionContext,
)
from dagster.utils.compatability import (
    create_custom_source_input, create_single_materialization_output
)


def silencing_default_context():
    return {
        'default':
        dagster.PipelineContextDefinition(
            argument_def_dict={},
            context_fn=lambda _pipeline, _args: ExecutionContext(),
        )
    }


def silencing_pipeline(solids):
    return dagster.PipelineDefinition(
        solids=solids, context_definitions=silencing_default_context()
    )


def create_failing_output_def():
    def failing_materialization_fn(*_args, **_kwargs):
        raise Exception('something bad happened')

    return create_single_materialization_output(
        name='CUSTOM',
        materialization_fn=failing_materialization_fn,
        argument_def_dict={},
    )


def create_input_set_input_def(input_name):
    return create_custom_source_input(
        input_name,
        source_fn=lambda context, arg_dict: [{input_name: 'input_set'}],
        argument_def_dict={},
    )


def create_root_success_solid(name):
    input_name = name + '_input'

    def root_transform(_context, args):
        passed_rows = list(args.values())[0]
        passed_rows.append({name: 'transform_called'})
        return passed_rows

    return SolidDefinition(
        name=name,
        inputs=[create_input_set_input_def(input_name)],
        transform_fn=root_transform,
        output=dagster.OutputDefinition(),
    )


def create_root_transform_failure_solid(name):
    input_name = name + '_input'
    inp = create_custom_source_input(
        input_name,
        source_fn=lambda context, arg_dict: [{input_name: 'input_set'}],
        argument_def_dict={},
    )

    def failed_transform(**_kwargs):
        raise Exception('Transform failed')

    return SolidDefinition(
        name=name,
        inputs=[inp],
        transform_fn=failed_transform,
        output=dagster.OutputDefinition(),
    )


def create_root_input_failure_solid(name):
    def failed_input_fn(_context, _args):
        raise Exception('something bad happened')

    input_name = name + '_input'
    inp = create_custom_source_input(
        input_name,
        source_fn=failed_input_fn,
        argument_def_dict={},
    )

    return SolidDefinition(
        name=name,
        inputs=[inp],
        transform_fn=lambda **_kwargs: {},
        output=dagster.OutputDefinition(),
    )


def create_root_output_failure_solid(name):
    input_name = name + '_input'

    def root_transform(**kwargs):
        passed_rows = list(kwargs.values())[0]
        passed_rows.append({name: 'transform_called'})
        return passed_rows

    return SolidDefinition(
        name=name,
        inputs=[create_input_set_input_def(input_name)],
        transform_fn=root_transform,
        output=create_failing_output_def(),
    )


def no_args_env(solid_name, input_name, materializations=None):
    return config.Environment(
        sources={solid_name: {
            input_name: config.Source(name='CUSTOM', args={})
        }},
        materializations=materializations,
    )


def test_transform_failure_pipeline():
    pipeline = silencing_pipeline(solids=[create_root_transform_failure_solid('failing')])
    pipeline_result = execute_pipeline(
        pipeline, environment=no_args_env('failing', 'failing_input'), throw_on_error=False
    )

    assert not pipeline_result.success

    result_list = pipeline_result.result_list

    assert len(result_list) == 1
    assert not result_list[0].success
    assert result_list[0].exception


def test_input_failure_pipeline():
    pipeline = silencing_pipeline(solids=[create_root_input_failure_solid('failing_input')])
    pipeline_result = execute_pipeline(
        pipeline,
        environment=no_args_env('failing_input', 'failing_input_input'),
        throw_on_error=False
    )

    result_list = pipeline_result.result_list

    assert len(result_list) == 1
    assert not result_list[0].success
    assert result_list[0].exception


def test_output_failure_pipeline():
    pipeline = silencing_pipeline(solids=[create_root_output_failure_solid('failing_output')])

    pipeline_result = execute_pipeline(
        pipeline,
        environment=no_args_env(
            'failing_output', 'failing_output_input',
            [config.Materialization(solid='failing_output', name='CUSTOM', args={})]
        ),
        throw_on_error=False,
    )

    assert not pipeline_result.success

    result_list = pipeline_result.result_list

    assert len(result_list) == 1
    assert not result_list[0].success
    assert result_list[0].exception


def test_failure_midstream():
    solid_a = create_root_success_solid('A')
    solid_b = create_root_success_solid('B')

    def transform_fn(_context, args):
        check.failed('user error')
        return [args['A'], args['B'], {'C': 'transform_called'}]

    solid_c = SolidDefinition(
        name='C',
        inputs=[
            InputDefinition(name='A', depends_on=solid_a),
            InputDefinition(name='B', depends_on=solid_b),
        ],
        transform_fn=transform_fn,
        output=dagster.OutputDefinition(),
    )

    environment = config.Environment(
        sources={
            'A': {
                'A_input': config.Source('CUSTOM', {}),
            },
            'B': {
                'B_input': config.Source('CUSTOM', {})
            }
        }
    )
    pipeline = silencing_pipeline(solids=[solid_a, solid_b, solid_c])
    pipeline_result = execute_pipeline(
        pipeline,
        environment=environment,
        throw_on_error=False,
    )

    result_list = pipeline_result.result_list

    assert result_list[0].success
    assert result_list[1].success
    assert not result_list[2].success
    assert result_list[2].reason == DagsterExecutionFailureReason.USER_CODE_ERROR



import dagster
from dagster import config, InputDefinition
from dagster.core.execution import execute_single_solid
from dagster.sqlalchemy_kernel.subquery_builder_experimental import (
    create_sql_statement_solid, sql_file_solid
)
from dagster.utils.test import script_relative_path

from .math_test_db import in_mem_context


def pipeline_test_def(solids, context):
    return dagster.PipelineDefinition(
        solids=solids,
        context_definitions={
            'default':
            dagster.PipelineContextDefinition(
                argument_def_dict={},
                context_fn=lambda _pipeline, _args: context,
            ),
        }
    )


def test_basic_isolated_sql_solid():
    context = in_mem_context()

    sql_text = '''CREATE TABLE sum_table AS SELECT num1, num2, num1 + num2 as sum FROM num_table'''

    basic_isolated_sql_solid = create_sql_statement_solid('basic_isolated_sql_solid', sql_text)

    result = execute_single_solid(
        context, basic_isolated_sql_solid, environment=config.Environment.empty()
    )

    assert result.success

    results = context.resources.sa.engine.connect().execute('SELECT * from sum_table').fetchall()
    assert results == [(1, 2, 3), (3, 4, 7)]


def test_basic_pipeline():
    sum_sql_text = '''CREATE TABLE sum_table AS
            SELECT num1, num2, num1 + num2 as sum FROM num_table'''

    sum_sq_sql_text = '''CREATE TABLE sum_sq_table AS
            SELECT num1, num2, sum, sum * sum as sum_sq FROM sum_table'''

    sum_sql_solid = create_sql_statement_solid('sum_sql_solid', sum_sql_text)

    sum_sq_sql_solid = create_sql_statement_solid(
        'sum_sq_sql_solid',
        sum_sq_sql_text,
        inputs=[InputDefinition(name=sum_sql_solid.name, depends_on=sum_sql_solid)]
    )

    pipeline = pipeline_test_def(solids=[sum_sql_solid, sum_sq_sql_solid], context=in_mem_context())

    pipeline_result = dagster.execute_pipeline(pipeline, environment=config.Environment.empty())

    assert pipeline_result.success

    exec_results = pipeline_result.result_list

    assert len(exec_results) == 2

    for exec_result in exec_results:
        assert exec_result.success is True

    engine = pipeline_result.context.resources.sa.engine

    results = engine.connect().execute('SELECT * from sum_table').fetchall()
    assert results == [(1, 2, 3), (3, 4, 7)]

    results = engine.connect().execute('SELECT * from sum_sq_table').fetchall()
    assert results == [(1, 2, 3, 9), (3, 4, 7, 49)]


def test_pipeline_from_files():
    create_sum_table_solid = sql_file_solid(script_relative_path('sql_files/create_sum_table.sql'))

    create_sum_sq_table_solid = sql_file_solid(
        script_relative_path('sql_files/create_sum_sq_table.sql'),
        inputs=[InputDefinition(create_sum_table_solid.name, depends_on=create_sum_table_solid)],
    )

    pipeline = pipeline_test_def(
        solids=[create_sum_table_solid, create_sum_sq_table_solid], context=in_mem_context()
    )

    pipeline_result = dagster.execute_pipeline(pipeline, environment=config.Environment.empty())

    assert pipeline_result.success

    exec_results = pipeline_result.result_list

    for exec_result in exec_results:
        assert exec_result.success is True

    engine = pipeline_result.context.resources.sa.engine

    results = engine.connect().execute('SELECT * from sum_table').fetchall()
    assert results == [(1, 2, 3), (3, 4, 7)]

    results = engine.connect().execute('SELECT * from sum_sq_table').fetchall()
    assert results == [(1, 2, 3, 9), (3, 4, 7, 49)]



from contextlib import contextmanager

from dagster import check

class IndentingPrinter:
    def __init__(self, indent_level=2, printer=print):
        self.lines = []
        self.current_indent = 0
        self.indent_level = check.int_param(indent_level, 'indent_level')
        self.printer = check.callable_param(printer, 'printer')

    def line(self, text):
        check.str_param(text, 'text')
        self.printer(self.current_indent_str + text)

    @property
    def current_indent_str(self):
        return ' ' * self.current_indent

    def blank_line(self):
        self.printer('')

    def increase_indent(self):
        self.current_indent += self.indent_level

    def decrease_indent(self):
        if self.current_indent <= 0:
            raise Exception('indent cannot be negative')
        self.current_indent -= self.indent_level

    @contextmanager
    def with_indent(self, text=None):
        if text is not None:
            self.line(text)
        self.increase_indent()
        yield
        self.decrease_indent()



from __future__ import (absolute_import, division, print_function, unicode_literals)
from builtins import *  # pylint: disable=W0622,W0401

from dagster.core.execution import (
    execute_pipeline,
    ExecutionContext,
    ExpectationResult,
    InputExpectationInfo,
    OutputExpectationInfo,
    InputExpectationResult,
    InputExpectationResults,
)

from dagster.core.definitions import (
    PipelineContextDefinition,
    PipelineDefinition,
    InputDefinition,
    OutputDefinition,
    MaterializationDefinition,
    ExpectationDefinition,
    SourceDefinition,
    SolidDefinition,
    ArgumentDefinition,
)

from dagster.core.decorators import (
    solid,
    source,
    materialization,
    with_context,
)

import dagster.core.types as types



# pylint: disable=W0613

import pytest

import dagster
import dagster.pandas_kernel as dagster_pd
from dagster import (config, InputDefinition, OutputDefinition, SolidDefinition)
from dagster.core.decorators import solid, source
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.execution import execute_single_solid
from dagster.utils.test import script_relative_path


def _dataframe_solid(name, inputs, transform_fn):
    return SolidDefinition(
        name=name,
        inputs=inputs,
        transform_fn=transform_fn,
        output=OutputDefinition(dagster_pd.DataFrame),
    )


def test_wrong_output_value():
    csv_input = InputDefinition('num_csv', dagster_pd.DataFrame)

    @solid(
        name="test_wrong_output", inputs=[csv_input], output=OutputDefinition(dagster_pd.DataFrame)
    )
    def df_solid(num_csv):
        return 'not a dataframe'

    with pytest.raises(DagsterInvariantViolationError):
        execute_single_solid(
            dagster.ExecutionContext(),
            df_solid,
            environment=config.Environment(
                sources={
                    'test_wrong_output': {
                        'num_csv': config.Source('CSV', {'path': script_relative_path('num.csv')})
                    },
                },
            ),
        )


def test_wrong_input_value():
    @source(name="WRONG")
    def wrong_source():
        return 'not a dataframe'

    @solid(
        name="test_wrong_input",
        inputs=[InputDefinition('foo', dagster_pd.DataFrame, sources=[wrong_source])],
        output=OutputDefinition(),
    )
    def df_solid(foo):
        return foo

    with pytest.raises(DagsterInvariantViolationError):
        execute_single_solid(
            dagster.ExecutionContext(),
            df_solid,
            environment=config.Environment(
                sources={'test_wrong_input': {
                    'foo': config.Source('WRONG', {})
                }}
            )
        )


def test_wrong_input_arg_dict():
    csv_input = InputDefinition('num_csv', dagster_pd.DataFrame)

    def transform_fn(context, args):
        return args['num_csv']

    df_solid = _dataframe_solid(
        name='test_wrong_value', inputs=[csv_input], transform_fn=transform_fn
    )

    with pytest.raises(DagsterInvariantViolationError):
        execute_single_solid(
            dagster.ExecutionContext(),
            df_solid,
            environment=config.Environment(
                sources={
                    'test_wrong_value': {
                        'num_jdkfjskdfjs':
                        config.Source('CSV', {'path': script_relative_path('num.csv')})
                    },
                },
            ),
        )



from dagster import check
from .errors import DagsterTypeError
from .definitions import check_argument_def_dict


def validate_args(argument_def_dict, arg_dict, error_context_str):
    check_argument_def_dict(argument_def_dict)
    check.dict_param(arg_dict, 'arg_dict', key_type=str)
    check.str_param(error_context_str, 'error_context_str')

    defined_args = set(argument_def_dict.keys())
    received_args = set(arg_dict.keys())

    for received_arg in received_args:
        if received_arg not in defined_args:
            raise DagsterTypeError(
                'Argument {received} not found in {error_context_str}. Defined args: {defined}'.
                format(
                    error_context_str=error_context_str,
                    defined=repr(defined_args),
                    received=received_arg,
                )
            )

    for expected_arg, arg_def in argument_def_dict.items():
        if arg_def.is_optional:
            continue

        check.invariant(not arg_def.default_provided)

        if expected_arg not in received_args:
            raise DagsterTypeError(
                'Did not not find {expected} in {error_context_str}. Defined args: {defined}'.
                format(
                    error_context_str=error_context_str,
                    expected=expected_arg,
                    defined=repr(defined_args),
                )
            )

    args_to_pass = {}

    for expected_arg, arg_def in argument_def_dict.items():
        if expected_arg in received_args:
            args_to_pass[expected_arg] = arg_dict[expected_arg]
        elif arg_def.default_provided:
            args_to_pass[expected_arg] = arg_def.default_value
        else:
            check.invariant(arg_def.is_optional and not arg_def.default_provided)

    for arg_name, arg_value in arg_dict.items():
        arg_def = argument_def_dict[arg_name]
        if not arg_def.dagster_type.is_python_valid_value(arg_value):
            format_string = (
                'Expected type {typename} for arg {arg_name} ' +
                'for {error_context_str} but got type "{arg_type}" value {arg_value}'
            )
            raise DagsterTypeError(
                format_string.format(
                    typename=arg_def.dagster_type.name,
                    arg_name=arg_name,
                    error_context_str=error_context_str,
                    arg_type=type(arg_value).__name__,
                    arg_value=repr(arg_value),
                )
            )

    return args_to_pass



import sqlalchemy

from dagster import check
from dagster.core.execution import ExecutionContext


class SqlAlchemyResource:
    def __init__(self, engine, mock_sql=False):
        self.engine = check.inst_param(engine, 'engine', sqlalchemy.engine.Engine)
        self.mock_sql = check.bool_param(mock_sql, 'mock_sql')


class DefaultSqlAlchemyResources:
    def __init__(self, sa):
        self.sa = check.inst_param(sa, 'sa', SqlAlchemyResource)


def check_supports_sql_alchemy_resource(context):
    check.inst_param(context, 'context', ExecutionContext)
    check.invariant(context.resources is not None)
    check.invariant(
        hasattr(context.resources, 'sa'),
        'Resources must have sa property be an object of SqlAlchemyResource',
    )
    check.inst(
        context.resources.sa,
        SqlAlchemyResource,
        'Resources must have sa property be an object of SqlAlchemyResource',
    )
    return context


def create_sql_alchemy_context_from_sa_resource(sa_resource, *args, **kwargs):
    check.inst_param(sa_resource, 'sa_resource', SqlAlchemyResource)
    resources = DefaultSqlAlchemyResources(sa_resource)
    context = ExecutionContext(resources=resources, *args, **kwargs)
    return check_supports_sql_alchemy_resource(context)


def create_sql_alchemy_context_from_engine(engine, *args, **kwargs):
    resources = DefaultSqlAlchemyResources(SqlAlchemyResource(engine))
    context = ExecutionContext(resources=resources, *args, **kwargs)
    return check_supports_sql_alchemy_resource(context)


def _is_sqlite_context(context):
    check_supports_sql_alchemy_resource(context)
    raw_connection = context.resources.sa.engine.raw_connection()
    if not hasattr(raw_connection, 'connection'):
        return False

    # FIXME: This is an awful way to do this
    return type(raw_connection.connection).__module__ == 'sqlite3'


def execute_sql_text_on_context(context, sql_text):
    check_supports_sql_alchemy_resource(context)
    check.str_param(sql_text, 'sql_text')

    if context.resources.sa.mock_sql:
        return

    engine = context.resources.sa.engine

    if _is_sqlite_context(context):
        # sqlite3 does not support multiple statements in a single
        # sql text and sqlalchemy does not abstract that away AFAICT
        # so have to hack around this
        raw_connection = engine.raw_connection()
        cursor = raw_connection.cursor()
        try:
            cursor.executescript(sql_text)
            raw_connection.commit()
        finally:
            cursor.close()
    else:
        connection = engine.connect()
        transaction = connection.begin()
        connection.execute(sql_text)
        transaction.commit()



import inspect
from functools import wraps
from dagster import check
from .definitions import (
    SolidDefinition,
    SourceDefinition,
    InputDefinition,
    OutputDefinition,
    MaterializationDefinition,
    DagsterInvalidDefinitionError,
)

# Error messages are long
# pylint: disable=C0301


def with_context(fn):
    """Pass context as a first argument to a transform, source or materialization function.
    """
    return _WithContext(fn)


class _WithContext:
    def __init__(self, fn):
        self.fn = fn

    @property
    def has_context(self):
        return True


class _Solid:
    def __init__(self, name=None, inputs=None, output=None, description=None):
        self.name = check.opt_str_param(name, 'name')
        self.inputs = check.opt_list_param(inputs, 'inputs', InputDefinition)

        check.opt_inst_param(output, 'output', OutputDefinition)
        if not output:
            output = OutputDefinition()
        self.output = output
        self.description = check.opt_str_param(description, 'description')

    def __call__(self, fn):
        expect_context = getattr(fn, 'has_context', False)
        if expect_context:
            fn = fn.fn

        if not self.name:
            self.name = fn.__name__

        _validate_transform_fn(self.name, fn, self.inputs, expect_context)
        transform_fn = _create_transform_wrapper(fn, self.inputs, expect_context)
        return SolidDefinition(
            name=self.name,
            inputs=self.inputs,
            output=self.output,
            transform_fn=transform_fn,
            description=self.description,
        )


def solid(*, name=None, inputs=None, output=None, description=None):
    return _Solid(name=name, inputs=inputs, output=output, description=description)


def _create_transform_wrapper(fn, inputs, include_context=False):
    input_names = [input.name for input in inputs]

    @wraps(fn)
    def transform(context, args):
        kwargs = {}
        for input_name in input_names:
            kwargs[input_name] = args[input_name]

        if include_context:
            return fn(context, **kwargs)
        else:
            return fn(**kwargs)

    return transform


class _Source:
    def __init__(self, name=None, argument_def_dict=None, description=None):
        self.source_type = check.opt_str_param(name, 'name')
        self.argument_def_dict = check.opt_dict_param(argument_def_dict, 'argument_def_dict')
        self.description = check.opt_str_param(description, 'description')

    def __call__(self, fn):
        include_context = getattr(fn, 'has_context', False)
        if include_context:
            fn = fn.fn

        if not self.source_type:
            self.source_type = fn.__name__

        _validate_source_fn(fn, self.source_type, self.argument_def_dict, include_context)
        source_fn = _create_source_wrapper(fn, self.argument_def_dict, include_context)

        return SourceDefinition(
            source_type=self.source_type,
            source_fn=source_fn,
            argument_def_dict=self.argument_def_dict,
            description=self.description,
        )


def source(*, name=None, argument_def_dict=None, description=None):
    return _Source(name=name, argument_def_dict=argument_def_dict, description=description)


def _create_source_wrapper(fn, arg_def_dict, include_context=False):
    arg_names = arg_def_dict.keys()

    @wraps(fn)
    def source_fn(context, args):
        kwargs = {}
        for arg in arg_names:
            kwargs[arg] = args[arg]

        if include_context:
            return fn(context, **kwargs)
        else:
            return fn(**kwargs)

    return source_fn


class _Materialization:
    def __init__(self, name=None, argument_def_dict=None, description=None):
        self.name = check.opt_str_param(name, 'name')
        self.argument_def_dict = check.opt_dict_param(argument_def_dict, 'argument_def_dict')
        self.description = check.opt_str_param(description, 'description')

    def __call__(self, fn):
        include_context = getattr(fn, 'has_context', False)
        if include_context:
            fn = fn.fn

        if not self.name:
            self.name = fn.__name__

        _validate_materialization_fn(fn, self.name, self.argument_def_dict, include_context)
        materialization_fn = _create_materialization_wrapper(
            fn, self.argument_def_dict, include_context
        )

        return MaterializationDefinition(
            name=self.name,
            materialization_fn=materialization_fn,
            argument_def_dict=self.argument_def_dict,
            description=self.description,
        )


def materialization(*, name=None, argument_def_dict=None, description=None):
    return _Materialization(
        name=name,
        argument_def_dict=argument_def_dict,
        description=description,
    )


def _create_materialization_wrapper(fn, arg_def_dict, include_context=False):
    arg_names = arg_def_dict.keys()

    @wraps(fn)
    def materialization_fn(context, args, data):
        kwargs = {}
        for arg in arg_names:
            kwargs[arg] = args[arg]

        if include_context:
            return fn(context, data, **kwargs)
        else:
            return fn(data, **kwargs)

    return materialization_fn


class FunctionValidationError(Exception):
    TYPES = {
        'vararg': 1,
        'missing_name': 2,
        'missing_positional': 3,
        'extra': 4,
    }

    def __init__(self, error_type, param=None, missing_names=None, **kwargs):
        super().__init__(**kwargs)
        self.error_type = error_type
        self.param = param
        self.missing_names = missing_names


def _validate_transform_fn(solid_name, transform_fn, inputs, expect_context=False):
    names = set(inp.name for inp in inputs)
    if expect_context:
        expected_positionals = ('context', )
    else:
        expected_positionals = ()
    try:
        _validate_decorated_fn(transform_fn, names, expected_positionals)
    except FunctionValidationError as e:
        if e.error_type == FunctionValidationError.TYPES['vararg']:
            raise DagsterInvalidDefinitionError(
                f"solid '{solid_name}' transform function has positional vararg parameter '{e.param}'. Transform functions should only have keyword arguments that match input names and optionally a first positional parameter named 'context'."
            )
        elif e.error_type == FunctionValidationError.TYPES['missing_name']:
            raise DagsterInvalidDefinitionError(
                f"solid '{solid_name}' transform function has parameter '{e.param}' that is not one of the solid inputs. Transform functions should only have keyword arguments that match input names and optionally a first positional parameter named 'context'."
            )
        elif e.error_type == FunctionValidationError.TYPES['missing_positional']:
            raise DagsterInvalidDefinitionError(
                f"solid '{solid_name}' transform function do not have required positional parameter '{e.param}'. Transform functions should only have keyword arguments that match input names and optionally a first positional parameter named 'context'."
            )
        elif e.error_type == FunctionValidationError.TYPES['extra']:
            undeclared_inputs_printed = ", '".join(e.missing_names)
            raise DagsterInvalidDefinitionError(
                f"solid '{solid_name}' transform function do not have parameter(s) '{undeclared_inputs_printed}', which are in solid's inputs. Transform functions should only have keyword arguments that match input names and optionally a first positional parameter named 'context'."
            )
        else:
            raise e


def _validate_source_fn(fn, source_name, arg_def_dict, expect_context=False):
    names = set(arg_def_dict.keys())
    if expect_context:
        expected_positionals = ('context', )
    else:
        expected_positionals = ()
    try:
        _validate_decorated_fn(fn, names, expected_positionals)
    except FunctionValidationError as e:
        if e.error_type == FunctionValidationError.TYPES['vararg']:
            raise DagsterInvalidDefinitionError(
                f"source '{source_name}' source function has positional vararg parameter '{e.param}'. Source functions should only have keyword arguments that match argument definition names and optionally a first positional parameter named 'context'."
            )
        elif e.error_type == FunctionValidationError.TYPES['missing_name']:
            raise DagsterInvalidDefinitionError(
                f"source '{source_name}' source function has parameter '{e.param}' that is not one of the source arguments. Source functions should only have keyword arguments that match argument definition names and optionally a first positional parameter named 'context'."
            )
        elif e.error_type == FunctionValidationError.TYPES['missing_positional']:
            raise DagsterInvalidDefinitionError(
                f"source '{source_name}' transform function do not have required positional argument '{e.param}'. Source functions should only have keyword arguments that match argument definition names and optionally a first positional parameter named 'context"
            )
        elif e.error_type == FunctionValidationError.TYPES['extra']:
            undeclared_inputs_printed = ", '".join(e.missing_names)
            raise DagsterInvalidDefinitionError(
                f"source '{source_name}' transform function do not have parameter(s) '{undeclared_inputs_printed}', which are in source's argument definitions.  Source functions should only have keyword arguments that match argument definition names and optionally a first positional parameter named 'context'"
            )
        else:
            raise e


def _validate_materialization_fn(fn, materialization_name, arg_def_dict, expect_context=False):
    names = set(arg_def_dict.keys())
    if expect_context:
        expected_positionals = ('context', 'data')
    else:
        expected_positionals = ('data', )
    try:
        _validate_decorated_fn(fn, names, expected_positionals)
    except FunctionValidationError as e:
        if e.error_type == FunctionValidationError.TYPES['vararg']:
            raise DagsterInvalidDefinitionError(
                f"materialization '{materialization_name}' materialization function has positional vararg parameter '{e.param}'. Materialization functions should only have keyword arguments that match argument definition names, positional parameter 'data' and optionally a first positional parameter named 'context'."
            )
        elif e.error_type == FunctionValidationError.TYPES['missing_name']:
            raise DagsterInvalidDefinitionError(
                f"materialization '{materialization_name}' transform function has parameter '{e.param}' that is not one of the materialization's argument definitions. Materialization functions should only have keyword arguments that match argument definition names, positional parameter 'data' and optionally a first positional parameter named 'context'."
            )
        elif e.error_type == FunctionValidationError.TYPES['missing_positional']:
            raise DagsterInvalidDefinitionError(
                f"materialization '{materialization_name}' materialization function do not have parameter {e.param}'. Materialization functions should only have keyword arguments that match argument definition names, positional parameter 'data' and optionally a first positional parameter named 'context'."
            )
        elif e.error_type == FunctionValidationError.TYPES['extra']:
            undeclared_inputs_printed = ", '".join(e.missing_names)
            raise DagsterInvalidDefinitionError(
                f"materialization '{materialization_name}' materialization function do not have parameter(s) '{undeclared_inputs_printed}', which are in materialization's argument definitinio. Materialization functions should only have keyword arguments that match argument definition names, positional parameter 'data' and optionally a first positional parameter named 'context'."
            )
        else:
            raise e


def _validate_decorated_fn(fn, names, expected_positionals):
    used_inputs = set()
    has_kwargs = False

    signature = inspect.signature(fn)
    params = list(signature.parameters.values())

    expected_positional_params = params[0:len(expected_positionals)]
    other_params = params[len(expected_positionals):]

    for expected, actual in zip(expected_positionals, expected_positional_params):
        possible_names = [expected, f'_{expected}', f'{expected}_']
        if (
            actual.kind not in [
                inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.POSITIONAL_ONLY
            ]
        ) or (actual.name not in possible_names):
            raise FunctionValidationError(
                FunctionValidationError.TYPES['missing_positional'], param=expected
            )

    for param in other_params:
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            has_kwargs = True
        elif param.kind == inspect.Parameter.VAR_POSITIONAL:
            raise FunctionValidationError(error_type=FunctionValidationError.TYPES['vararg'])

        else:
            if param.name not in names:
                raise FunctionValidationError(
                    FunctionValidationError.TYPES['missing_name'], param=param.name
                )
            else:
                used_inputs.add(param.name)

    undeclared_inputs = names - used_inputs
    if not has_kwargs and undeclared_inputs:
        raise FunctionValidationError(
            FunctionValidationError.TYPES['extra'], missing_names=undeclared_inputs
        )



import os

from dagster import config

from dagster.core.execution import execute_pipeline
from dagster.utils.test import script_relative_path
from dagster.cli.pipeline import (do_execute_command, print_pipeline)

from dagster.dagster_examples.pandas_hello_world.pipeline import define_pipeline


def test_pipeline_include():
    assert define_pipeline()


def test_execute_pipeline():
    pipeline = define_pipeline()
    environment = config.Environment(
        sources={
            'sum_solid': {
                'num': config.Source(name='CSV', args={'path': script_relative_path('num.csv')})
            }
        },
        execution=config.Execution(from_solids=['sum_solid'], through_solids=['sum_sq_solid']),
    )

    result = execute_pipeline(pipeline, environment=environment)

    assert result.success

    assert result.result_named('sum_solid').transformed_value.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
    }

    assert result.result_named('sum_sq_solid').transformed_value.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
        'sum_sq': [9, 49],
    }


def test_cli_execute():

    # currently paths in env files have to be relative to where the
    # script has launched so we have to simulate that
    cwd = os.getcwd()
    try:

        os.chdir(script_relative_path('../..'))

        do_execute_command(
            define_pipeline(),
            script_relative_path('../../pandas_hello_world/env.yml'),
            lambda *_args, **_kwargs: None,
        )
    finally:
        # restore cwd
        os.chdir(cwd)


def test_cli_print():
    print_pipeline(define_pipeline(), full=False, print_fn=lambda *_args, **_kwargs: None)
    print_pipeline(define_pipeline(), full=True, print_fn=lambda *_args, **_kwargs: None)



from __future__ import (absolute_import, division, print_function, unicode_literals)
from builtins import *  # pylint: disable=W0622,W0401

import pandas as pd

import dagster
from dagster import (check, types)


def parquet_dataframe_source(**read_parquet_kwargs):
    def callback(context, arg_dict):
        check.inst_param(context, 'context', dagster.ExecutionContext)
        check.str_param(arg_dict['path'], 'path')
        df = pd.read_parquet(arg_dict['path'], **read_parquet_kwargs)
        context.metric('rows', df.shape[0])
        return df

    return dagster.SourceDefinition(
        source_type='PARQUET',
        source_fn=callback,
        argument_def_dict={
            'path': dagster.ArgumentDefinition(dagster.types.Path),
        },
    )


def csv_dataframe_source(name=None, **read_csv_kwargs):
    def callback(context, arg_dict):
        check.inst_param(context, 'context', dagster.ExecutionContext)
        check.str_param(arg_dict['path'], 'path')
        df = pd.read_csv(arg_dict['path'], **read_csv_kwargs)
        context.metric('rows', df.shape[0])
        return df

    return dagster.SourceDefinition(
        source_type=check.opt_str_param(name, 'name', 'CSV'),
        source_fn=callback,
        argument_def_dict={
            'path': dagster.ArgumentDefinition(dagster.types.Path),
        },
    )


def table_dataframe_source(**read_table_kwargs):
    def callback(context, arg_dict):
        check.inst_param(context, 'context', dagster.ExecutionContext)
        path = check.str_elem(arg_dict, 'path')
        df = pd.read_table(path, **read_table_kwargs)
        context.metric('rows', df.shape[0])
        return df

    return dagster.SourceDefinition(
        source_type='TABLE',
        source_fn=callback,
        argument_def_dict={
            'path': dagster.ArgumentDefinition(dagster.types.Path),
        },
    )


def dataframe_csv_materialization():
    def to_csv_fn(context, arg_dict, df):
        check.inst_param(df, 'df', pd.DataFrame)
        check.inst_param(context, 'context', dagster.ExecutionContext)
        check.dict_param(arg_dict, 'arg_dict')
        path = check.str_elem(arg_dict, 'path')

        df.to_csv(path, index=False)

    return dagster.MaterializationDefinition(
        name='CSV',
        materialization_fn=to_csv_fn,
        argument_def_dict={'path': dagster.ArgumentDefinition(types.Path)},
    )


def dataframe_parquet_materialization():
    def to_parquet_fn(context, arg_dict, df):
        check.inst_param(df, 'df', pd.DataFrame)
        check.inst_param(context, 'context', dagster.ExecutionContext)
        check.dict_param(arg_dict, 'arg_dict')
        path = check.str_elem(arg_dict, 'path')

        df.to_parquet(path)

    return dagster.MaterializationDefinition(
        name='PARQUET',
        materialization_fn=to_parquet_fn,
        argument_def_dict={'path': dagster.ArgumentDefinition(types.Path)},
    )


def _create_dataframe_type():
    return dagster.types.PythonObjectType(
        name='PandasDataFrame',
        python_type=pd.DataFrame,
        default_sources=[
            parquet_dataframe_source(),
            csv_dataframe_source(),
            table_dataframe_source(),
        ],
        default_materializations=[
            dataframe_csv_materialization(),
            dataframe_parquet_materialization(),
        ]
    )


DataFrame = _create_dataframe_type()



import importlib

import click
import yaml

from dagster import (check, PipelineDefinition)


def define_config_class():
    class _Config:
        def __init__(self, pipeline_configs):
            self.pipeline_configs = pipeline_configs
            self.pipelines = None

        @staticmethod
        def from_file(filepath):
            with open(filepath, 'r') as ff:
                config = yaml.load(ff)

            pipeline_configs = [
                PipelineConfig(
                    module=check.str_elem(entry, 'module'), fn=check.str_elem(entry, 'fn')
                ) for entry in check.list_elem(config, 'pipelines')
            ]

            return Config(pipeline_configs=pipeline_configs)

        def create_pipelines(self):
            for pipeline_config in self.pipeline_configs:
                pipeline_config.create_pipeline()

            return self.pipeline_configs

        def get_pipeline(self, name):
            if not self.pipelines:
                self.create_pipelines()

            for pipeline_config in self.pipeline_configs:
                if pipeline_config.pipeline.name == name:
                    return pipeline_config

            check.failed(f'pipeline {name} not found')

    # This lets you ask cli commands to have this object extracted from context and
    # passed as first arg
    _Config.pass_object = click.make_pass_decorator(_Config)
    return _Config


Config = define_config_class()


def define_pipeline_config_class():
    class _PipelineConfig:
        def __init__(self, module, fn):
            self.module_name = module
            self.fn_name = fn
            self.module = None
            self.fn = None
            self.pipeline = None

        def create_pipeline(self):
            self.module = importlib.import_module(self.module_name)
            self.fn = getattr(self.module, self.fn_name)
            check.is_callable(self.fn)
            self.pipeline = check.inst(self.fn(), PipelineDefinition)
            return self.pipeline

    # This lets you ask cli commands to have this object extracted from context and
    # passed as first arg
    _PipelineConfig.pass_object = click.make_pass_decorator(_PipelineConfig)
    return _PipelineConfig


PipelineConfig = define_pipeline_config_class()



import pytest

from dagster.utils.timing import (time_execution_scope, TimerResult)


def test_basic_usage():
    with time_execution_scope() as timer_result:
        pass

    assert timer_result
    assert isinstance(timer_result.millis, float)
    assert isinstance(timer_result.seconds, float)


def test_wrong_usage():
    with pytest.raises(Exception):
        timer_result = TimerResult()
        assert timer_result.millis

    with pytest.raises(Exception):
        timer_result = TimerResult()
        assert timer_result.seconds



import os

import pandas as pd

import dagster
from dagster import (
    check, config, InputDefinition, OutputDefinition, SolidDefinition, ArgumentDefinition
)
from dagster.core import types
from dagster.core.decorators import solid
from dagster.core.execution import (
    InMemoryInputManager,
    ExecutionContext,
    _read_source,
    execute_pipeline_iterator,
    output_single_solid,
    _pipeline_solid_in_memory,
    execute_pipeline,
    execute_single_solid,
)
import dagster.pandas_kernel as dagster_pd
from dagster.utils.compatability import create_single_materialization_output
from dagster.utils.test import (get_temp_file_name, get_temp_file_names, script_relative_path)


def _dataframe_solid(name, inputs, transform_fn):
    return SolidDefinition(
        name=name,
        inputs=inputs,
        transform_fn=transform_fn,
        output=dagster.OutputDefinition(dagster_pd.DataFrame),
    )


def get_solid_transformed_value(context, solid_inst, environment):
    execution_result = execute_single_solid(
        context,
        solid_inst,
        environment=environment,
    )
    return execution_result.transformed_value


def get_num_csv_environment(solid_name, materializations=None, through_solids=None):
    return config.Environment(
        sources={
            solid_name: {
                'num_csv': config.Source('CSV', args={'path': script_relative_path('num.csv')})
            },
        },
        materializations=materializations,
        execution=config.Execution(through_solids=through_solids),
    )


def create_test_context():
    return ExecutionContext()


def test_pandas_input():
    csv_input = InputDefinition('num_csv', dagster_pd.DataFrame)
    df = _read_source(
        create_test_context(), csv_input.source_of_type('CSV'),
        {'path': script_relative_path('num.csv')}
    )

    assert isinstance(df, pd.DataFrame)
    assert df.to_dict('list') == {'num1': [1, 3], 'num2': [2, 4]}


def test_pandas_solid():
    csv_input = InputDefinition('num_csv', dagster_pd.DataFrame)

    def transform(_context, args):
        num_csv = args['num_csv']
        num_csv['sum'] = num_csv['num1'] + num_csv['num2']
        return num_csv

    test_output = {}

    def materialization_fn_inst(context, arg_dict, df):
        assert isinstance(df, pd.DataFrame)
        assert isinstance(context, ExecutionContext)
        assert isinstance(arg_dict, dict)

        test_output['df'] = df

    custom_output_def = create_single_materialization_output(
        name='CUSTOM',
        materialization_fn=materialization_fn_inst,
        argument_def_dict={},
    )

    single_solid = SolidDefinition(
        name='sum_table',
        inputs=[csv_input],
        transform_fn=transform,
        output=custom_output_def,
    )

    output_single_solid(
        create_test_context(),
        single_solid,
        environment=get_num_csv_environment('sum_table'),
        name='CUSTOM',
        arg_dict={},
    )

    assert test_output['df'].to_dict('list') == {'num1': [1, 3], 'num2': [2, 4], 'sum': [3, 7]}


def test_pandas_csv_to_csv():
    csv_input = InputDefinition('num_csv', dagster_pd.DataFrame)

    # just adding a second context arg to test that
    def transform(context, args):
        check.inst_param(context, 'context', dagster.core.execution.ExecutionContext)
        num_csv = args['num_csv']
        num_csv['sum'] = num_csv['num1'] + num_csv['num2']
        return num_csv

    def materialization_fn_inst(context, arg_dict, df):
        assert isinstance(context, ExecutionContext)
        path = check.str_elem(arg_dict, 'path')
        df.to_csv(path, index=False)

    csv_output_def = create_single_materialization_output(
        name='CSV',
        materialization_fn=materialization_fn_inst,
        argument_def_dict={'path': ArgumentDefinition(types.Path)}
    )

    solid_def = SolidDefinition(
        name='sum_table',
        inputs=[csv_input],
        transform_fn=transform,
        output=csv_output_def,
    )

    output_df = execute_transform_in_temp_csv_files(solid_def)

    assert output_df.to_dict('list') == {'num1': [1, 3], 'num2': [2, 4], 'sum': [3, 7]}


def execute_transform_in_temp_csv_files(solid_inst):
    with get_temp_file_name() as temp_file_name:
        result = output_single_solid(
            create_test_context(),
            solid_inst,
            environment=get_num_csv_environment(solid_inst.name),
            name='CSV',
            arg_dict={'path': temp_file_name},
        )

        assert result.success

        output_df = pd.read_csv(temp_file_name)
    return output_df


def create_sum_table():
    def transform(_context, args):
        num_csv = args['num_csv']
        check.inst_param(num_csv, 'num_csv', pd.DataFrame)
        num_csv['sum'] = num_csv['num1'] + num_csv['num2']
        return num_csv

    return _dataframe_solid(
        name='sum_table',
        inputs=[InputDefinition('num_csv', dagster_pd.DataFrame)],
        transform_fn=transform,
    )


@solid(
    inputs=[dagster.InputDefinition('num_csv', dagster_pd.DataFrame)],
    output=dagster.OutputDefinition(dagster_pd.DataFrame),
)
def sum_table(num_csv):
    check.inst_param(num_csv, 'num_csv', pd.DataFrame)
    num_csv['sum'] = num_csv['num1'] + num_csv['num2']
    return num_csv


@solid(
    inputs=[dagster.InputDefinition('sum_df', dagster_pd.DataFrame, depends_on=sum_table)],
    output=dagster.OutputDefinition(dagster_pd.DataFrame),
)
def sum_sq_table(sum_df):
    sum_df['sum_squared'] = sum_df['sum'] * sum_df['sum']
    return sum_df


@solid(
    inputs=[
        dagster.InputDefinition('sum_table_renamed', dagster_pd.DataFrame, depends_on=sum_table)
    ],
    output=dagster.OutputDefinition(dagster_pd.DataFrame),
)
def sum_sq_table_renamed_input(sum_table_renamed):
    sum_table_renamed['sum_squared'] = sum_table_renamed['sum'] * sum_table_renamed['sum']
    return sum_table_renamed


def create_sum_sq_table(sum_table_solid):
    def transform(_context, args):
        sum_df = args['sum_table']
        sum_df['sum_squared'] = sum_df['sum'] * sum_df['sum']
        return sum_df

    return _dataframe_solid(
        name='mult_table',
        inputs=[
            dagster.InputDefinition('sum_table', dagster_pd.DataFrame, depends_on=sum_table_solid)
        ],
        transform_fn=transform
    )


def test_pandas_csv_to_csv_better_api():
    output_df = execute_transform_in_temp_csv_files(create_sum_table())
    assert output_df.to_dict('list') == {'num1': [1, 3], 'num2': [2, 4], 'sum': [3, 7]}


def test_pandas_csv_to_csv_decorator_api():
    output_df = execute_transform_in_temp_csv_files(sum_table)
    assert output_df.to_dict('list') == {'num1': [1, 3], 'num2': [2, 4], 'sum': [3, 7]}


def test_pandas_csv_in_memory():
    df = get_solid_transformed_value(
        create_test_context(),
        create_sum_table(),
        get_num_csv_environment('sum_table'),
    )
    assert isinstance(df, pd.DataFrame)
    assert df.to_dict('list') == {'num1': [1, 3], 'num2': [2, 4], 'sum': [3, 7]}


def test_two_step_pipeline_in_memory():
    sum_table_solid = create_sum_table()
    mult_table_solid = create_sum_sq_table(sum_table_solid)
    context = create_test_context()
    df = get_solid_transformed_value(context, sum_table_solid, get_num_csv_environment('sum_table'))
    input_values = {'sum_table': df}
    input_manager = InMemoryInputManager(context, input_values)
    mult_df = _pipeline_solid_in_memory(
        context, input_manager, mult_table_solid, input_values
    ).transformed_value
    assert mult_df.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
        'sum_squared': [9, 49]
    }


def test_two_step_pipeline_in_memory_decorator_style():
    context = create_test_context()
    df = get_solid_transformed_value(context, sum_table, get_num_csv_environment('sum_table'))
    input_values = {'sum_df': df}
    mult_df = _pipeline_solid_in_memory(
        context, InMemoryInputManager(context, input_values), sum_sq_table, input_values
    ).transformed_value
    assert mult_df.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
        'sum_squared': [9, 49]
    }


def test_two_input_solid():
    def transform(_context, args):
        num_csv1 = args['num_csv1']
        num_csv2 = args['num_csv2']
        check.inst_param(num_csv1, 'num_csv1', pd.DataFrame)
        check.inst_param(num_csv2, 'num_csv2', pd.DataFrame)
        num_csv1['sum'] = num_csv1['num1'] + num_csv2['num2']
        return num_csv1

    two_input_solid = _dataframe_solid(
        name='two_input_solid',
        inputs=[
            InputDefinition('num_csv1', dagster_pd.DataFrame),
            InputDefinition('num_csv2', dagster_pd.DataFrame),
        ],
        transform_fn=transform,
    )

    environment = config.Environment(
        sources={
            'two_input_solid': {
                'num_csv1': config.Source('CSV', {'path': script_relative_path('num.csv')}),
                'num_csv2': config.Source('CSV', {'path': script_relative_path('num.csv')}),
            },
        },
    )

    df = get_solid_transformed_value(create_test_context(), two_input_solid, environment)
    assert isinstance(df, pd.DataFrame)
    assert df.to_dict('list') == {'num1': [1, 3], 'num2': [2, 4], 'sum': [3, 7]}


def test_no_transform_solid():
    num_table = _dataframe_solid(
        name='num_table',
        inputs=[InputDefinition('num_csv', dagster_pd.DataFrame)],
        transform_fn=lambda _context, args: args['num_csv'],
    )
    context = create_test_context()
    df = get_solid_transformed_value(context, num_table, get_num_csv_environment('num_table'))
    assert df.to_dict('list') == {'num1': [1, 3], 'num2': [2, 4]}


def create_diamond_pipeline():
    return dagster.PipelineDefinition(solids=list(create_diamond_dag()))


def create_diamond_dag():
    num_table_solid = _dataframe_solid(
        name='num_table',
        inputs=[InputDefinition('num_csv', dagster_pd.DataFrame)],
        transform_fn=lambda _context, args: args['num_csv'],
    )

    def sum_transform(_context, args):
        num_df = args['num_table']
        sum_df = num_df.copy()
        sum_df['sum'] = num_df['num1'] + num_df['num2']
        return sum_df

    sum_table_solid = _dataframe_solid(
        name='sum_table',
        inputs=[
            dagster.InputDefinition('num_table', dagster_pd.DataFrame, depends_on=num_table_solid)
        ],
        transform_fn=sum_transform,
    )

    def mult_transform(_context, args):
        num_table = args['num_table']
        mult_table = num_table.copy()
        mult_table['mult'] = num_table['num1'] * num_table['num2']
        return mult_table

    mult_table_solid = _dataframe_solid(
        name='mult_table',
        inputs=[
            dagster.InputDefinition('num_table', dagster_pd.DataFrame, depends_on=num_table_solid)
        ],
        transform_fn=mult_transform,
    )

    def sum_mult_transform(_context, args):
        sum_df = args['sum_table']
        mult_df = args['mult_table']
        sum_mult_table = sum_df.copy()
        sum_mult_table['mult'] = mult_df['mult']
        sum_mult_table['sum_mult'] = sum_df['sum'] * mult_df['mult']
        return sum_mult_table

    sum_mult_table_solid = _dataframe_solid(
        name='sum_mult_table',
        inputs=[
            dagster.InputDefinition('sum_table', dagster_pd.DataFrame, depends_on=sum_table_solid),
            dagster.InputDefinition(
                'mult_table', dagster_pd.DataFrame, depends_on=mult_table_solid
            ),
        ],
        transform_fn=sum_mult_transform,
    )

    return (num_table_solid, sum_table_solid, mult_table_solid, sum_mult_table_solid)


def test_diamond_dag_run():
    num_table_solid, sum_table_solid, mult_table_solid, sum_mult_table_solid = create_diamond_dag()

    context = create_test_context()

    num_table_df = get_solid_transformed_value(
        context,
        num_table_solid,
        get_num_csv_environment('num_table'),
    )
    assert num_table_df.to_dict('list') == {'num1': [1, 3], 'num2': [2, 4]}

    input_values = {'num_table': num_table_df}
    sum_df = _pipeline_solid_in_memory(
        context, InMemoryInputManager(context, input_values), sum_table_solid, input_values
    ).transformed_value

    assert sum_df.to_dict('list') == {'num1': [1, 3], 'num2': [2, 4], 'sum': [3, 7]}

    input_values = {'num_table': num_table_df}
    mult_df = _pipeline_solid_in_memory(
        context, InMemoryInputManager(context, input_values), mult_table_solid, input_values
    ).transformed_value

    assert mult_df.to_dict('list') == {'num1': [1, 3], 'num2': [2, 4], 'mult': [2, 12]}

    input_values = {'sum_table': sum_df, 'mult_table': mult_df}

    sum_mult_df = _pipeline_solid_in_memory(
        context, InMemoryInputManager(context, input_values), sum_mult_table_solid, input_values
    ).transformed_value

    assert sum_mult_df.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
        'mult': [2, 12],
        'sum_mult': [6, 84],
    }


def test_pandas_in_memory_diamond_pipeline():
    pipeline = create_diamond_pipeline()
    result = execute_pipeline(
        pipeline,
        environment=get_num_csv_environment('num_table', through_solids=['sum_mult_table'])
    )

    assert result.result_named('sum_mult_table').transformed_value.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
        'mult': [2, 12],
        'sum_mult': [6, 84],
    }


def test_pandas_output_csv_pipeline():
    with get_temp_file_name() as temp_file_name:
        pipeline = create_diamond_pipeline()
        environment = get_num_csv_environment(
            'num_table', [
                config.Materialization(
                    solid='sum_mult_table',
                    name='CSV',
                    args={'path': temp_file_name},
                )
            ]
        )

        for _result in execute_pipeline_iterator(pipeline=pipeline, environment=environment):
            pass

        assert os.path.exists(temp_file_name)
        output_df = pd.read_csv(temp_file_name)
        assert output_df.to_dict('list') == {
            'num1': [1, 3],
            'num2': [2, 4],
            'sum': [3, 7],
            'mult': [2, 12],
            'sum_mult': [6, 84],
        }


def _result_named(results, name):
    for result in results:
        if result.name == name:
            return result

    check.failed('could not find name')


def test_pandas_output_intermediate_csv_files():
    pipeline = create_diamond_pipeline()

    with get_temp_file_names(2) as temp_tuple:
        sum_file, mult_file = temp_tuple  # pylint: disable=E0632

        environment = get_num_csv_environment(
            'num_table', [
                config.Materialization(
                    solid='sum_table',
                    name='CSV',
                    args={'path': sum_file},
                ),
                config.Materialization(
                    solid='mult_table',
                    name='CSV',
                    args={'path': mult_file},
                ),
            ]
        )

        subgraph_one_result = execute_pipeline(pipeline, environment=environment)

        assert len(subgraph_one_result.result_list) == 4

        expected_sum = {
            'num1': [1, 3],
            'num2': [2, 4],
            'sum': [3, 7],
        }

        assert pd.read_csv(sum_file).to_dict('list') == expected_sum
        sum_table_result = subgraph_one_result.result_named('sum_table')
        assert sum_table_result.transformed_value.to_dict('list') == expected_sum

        expected_mult = {
            'num1': [1, 3],
            'num2': [2, 4],
            'mult': [2, 12],
        }
        assert pd.read_csv(mult_file).to_dict('list') == expected_mult
        mult_table_result = subgraph_one_result.result_named('mult_table')
        assert mult_table_result.transformed_value.to_dict('list') == expected_mult

        pipeline_result = execute_pipeline(
            pipeline,
            environment=config.Environment(
                sources={
                    'sum_mult_table': {
                        'sum_table': config.Source('CSV', {'path': sum_file}),
                        'mult_table': config.Source('CSV', {'path': mult_file}),
                    },
                },
                execution=config.Execution.single_solid('sum_mult_table'),
            ),
        )

        assert pipeline_result.success

        subgraph_two_result_list = pipeline_result.result_list

        assert len(subgraph_two_result_list) == 1
        output_df = subgraph_two_result_list[0].transformed_value
        assert output_df.to_dict('list') == {
            'num1': [1, 3],
            'num2': [2, 4],
            'sum': [3, 7],
            'mult': [2, 12],
            'sum_mult': [6, 84],
        }


def csv_materialization(solid_name, path):
    return config.Materialization(
        solid=solid_name,
        name='CSV',
        args={'path': path},
    )


def parquet_materialization(solid_name, path):
    return config.Materialization(
        solid=solid_name,
        name='PARQUET',
        args={'path': path},
    )


def test_pandas_output_intermediate_parquet_files():
    pipeline = create_diamond_pipeline()

    with get_temp_file_names(2) as temp_tuple:
        # false positive on pylint error
        sum_file, mult_file = temp_tuple  # pylint: disable=E0632
        pipeline_result = execute_pipeline(
            pipeline,
            environment=get_num_csv_environment(
                'num_table', [
                    parquet_materialization('sum_table', sum_file),
                    parquet_materialization('mult_table', mult_file),
                ]
            ),
        )

        assert pipeline_result.success

        expected_sum = {
            'num1': [1, 3],
            'num2': [2, 4],
            'sum': [3, 7],
        }

        assert pd.read_parquet(sum_file).to_dict('list') == expected_sum


def test_pandas_multiple_inputs():

    environment = config.Environment(
        sources={
            'double_sum': {
                'num_csv1': config.Source('CSV', {'path': script_relative_path('num.csv')}),
                'num_csv2': config.Source('CSV', {'path': script_relative_path('num.csv')}),
            },
        },
        execution=config.Execution(through_solids=['double_sum']),
    )

    def transform_fn(_context, args):
        return args['num_csv1'] + args['num_csv2']

    double_sum = _dataframe_solid(
        name='double_sum',
        inputs=[
            InputDefinition('num_csv1', dagster_pd.DataFrame),
            InputDefinition('num_csv2', dagster_pd.DataFrame),
        ],
        transform_fn=transform_fn
    )
    pipeline = dagster.PipelineDefinition(solids=[double_sum])

    output_df = execute_pipeline(
        pipeline,
        environment=environment,
        # solid_name='double_sum',
    ).result_list[0].transformed_value

    assert not output_df.empty

    assert output_df.to_dict('list') == {
        'num1': [2, 6],
        'num2': [4, 8],
    }


def test_pandas_multiple_outputs():
    with get_temp_file_names(2) as temp_tuple:
        # false positive on pylint error
        csv_file, parquet_file = temp_tuple  # pylint: disable=E0632
        pipeline = create_diamond_pipeline()

        for _result in execute_pipeline_iterator(
            pipeline=pipeline,
            environment=get_num_csv_environment(
                'num_table', [
                    csv_materialization('sum_mult_table', csv_file),
                    parquet_materialization('sum_mult_table', parquet_file),
                ]
            ),
        ):
            pass

        assert os.path.exists(csv_file)
        output_csv_df = pd.read_csv(csv_file)
        assert output_csv_df.to_dict('list') == {
            'num1': [1, 3],
            'num2': [2, 4],
            'sum': [3, 7],
            'mult': [2, 12],
            'sum_mult': [6, 84],
        }

        assert os.path.exists(parquet_file)
        output_parquet_df = pd.read_parquet(parquet_file)
        assert output_parquet_df.to_dict('list') == {
            'num1': [1, 3],
            'num2': [2, 4],
            'sum': [3, 7],
            'mult': [2, 12],
            'sum_mult': [6, 84],
        }


def test_rename_input():
    result = execute_pipeline(
        dagster.PipelineDefinition(solids=[sum_table, sum_sq_table_renamed_input]),
        environment=get_num_csv_environment('sum_table'),
    )

    assert result.success

    assert result.result_named('sum_sq_table_renamed_input').transformed_value.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
        'sum_squared': [9, 49],
    }



import sys
import os

from setuptools import find_packages, setup

# pylint: disable=E0401, W0611
if sys.version_info[0] < 3:
    import __builtin__ as builtins
else:
    import builtins


def long_description():
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, 'README.rst'), 'r') as fh:
        return fh.read()


setup(
    name='dagster',
    version='0.1.5',
    author='Elementl',
    author_email='schrockn@elementl.com',
    license='Apache-2.0',
    description='Dagster is an opinionated pipeline runner.',
    long_description=long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/dagster-io/dagster',
    classifiers=(
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ),
    packages=find_packages(exclude=['dagster_tests']),
    install_requires=[
        # standard python 2/3 compatability things
        'enum34>=1.1.6',
        'future>=0.16.0',

        # cli
        'click>=6.7',
        'coloredlogs>=10.0',
        'graphviz>=0.8.3',
        'pyyaml>=3.12',

        # core (not explicitly expressed atm)
        'six>=1.11.0',
        'toposort>=1.5',

        # pandas kernel
        'pandas>=0.22.0',
        'pyarrow>=0.8.0',

        # sqlalchemy kernel
        'sqlalchemy>=1.2.7',
        'jinja2>=2.8',

        # examples
        # unused and very, very heavy
        # 'great-expectations>=0.4.1',
        # 'requests>=2.18.4',

        # dev/test - Installed via dev-requirements.txt
        # 'pylint>=1.8.4',
        # 'pytest>=3.5.1',
        # 'recommonmark>=0.4.0',
        # 'rope>=0.10.7',
        # 'Sphinx>=1.7.5',
        # 'sphinx-autobuild>=0.7.1',
        # 'yapf>=0.22.0',
        # 'twine>=1.11.0',
        # 'pre-commit'>=1.10.1',
    ],
    scripts=['bin/dagster']
)



import pytest

from dagster import check
from dagster.check import (
    ParameterCheckError, ElementCheckError, CheckError, NotImplementedCheckError
)


def test_int_param():
    assert check.int_param(-1, 'param_name') == -1
    assert check.int_param(0, 'param_name') == 0
    assert check.int_param(1, 'param_name') == 1

    with pytest.raises(ParameterCheckError):
        check.int_param(None, 'param_name')

    with pytest.raises(ParameterCheckError):
        check.int_param('s', 'param_name')


def test_opt_int_param():
    assert check.opt_int_param(-1, 'param_name') == -1
    assert check.opt_int_param(0, 'param_name') == 0
    assert check.opt_int_param(1, 'param_name') == 1
    assert check.opt_int_param(None, 'param_name') is None

    with pytest.raises(ParameterCheckError):
        check.opt_int_param('s', 'param_name')


def test_list_param():
    assert check.list_param([], 'list_param') == []

    with pytest.raises(ParameterCheckError):
        check.list_param(None, 'list_param')

    with pytest.raises(ParameterCheckError):
        check.list_param('3u4', 'list_param')


def test_typed_list_param():
    class Foo:
        pass

    class Bar:
        pass

    assert check.list_param([], 'list_param', Foo) == []
    foo_list = [Foo()]
    assert check.list_param(foo_list, 'list_param', Foo) == foo_list

    with pytest.raises(CheckError):
        check.list_param([Bar()], 'list_param', Foo)

    with pytest.raises(CheckError):
        check.list_param([None], 'list_param', Foo)


def test_opt_list_param():
    assert check.opt_list_param(None, 'list_param') == []
    assert check.opt_list_param([], 'list_param') == []
    obj_list = [1]
    assert check.list_param(obj_list, 'list_param') == obj_list

    with pytest.raises(ParameterCheckError):
        check.opt_list_param(0, 'list_param')

    with pytest.raises(ParameterCheckError):
        check.opt_list_param('', 'list_param')

    with pytest.raises(ParameterCheckError):
        check.opt_list_param('3u4', 'list_param')


def test_opt_typed_list_param():
    class Foo:
        pass

    class Bar:
        pass

    assert check.opt_list_param(None, 'list_param', Foo) == []
    assert check.opt_list_param([], 'list_param', Foo) == []
    foo_list = [Foo()]
    assert check.opt_list_param(foo_list, 'list_param', Foo) == foo_list

    with pytest.raises(CheckError):
        check.opt_list_param([Bar()], 'list_param', Foo)

    with pytest.raises(CheckError):
        check.opt_list_param([None], 'list_param', Foo)


def test_dict_param():
    assert check.dict_param({}, 'dict_param') == {}
    ddict = {'a': 2}
    assert check.dict_param(ddict, 'dict_param') == ddict

    with pytest.raises(ParameterCheckError):
        check.dict_param(None, 'dict_param')

    with pytest.raises(ParameterCheckError):
        check.dict_param(0, 'dict_param')

    with pytest.raises(ParameterCheckError):
        check.dict_param(1, 'dict_param')

    with pytest.raises(ParameterCheckError):
        check.dict_param('foo', 'dict_param')

    with pytest.raises(ParameterCheckError):
        check.dict_param(['foo'], 'dict_param')

    with pytest.raises(ParameterCheckError):
        check.dict_param([], 'dict_param')


def test_dict_param_with_type():
    str_to_int = {'str': 1}
    assert check.dict_param(str_to_int, 'str_to_int', key_type=str, value_type=int)
    assert check.dict_param(str_to_int, 'str_to_int', value_type=int)
    assert check.dict_param(str_to_int, 'str_to_int', key_type=str)
    assert check.dict_param(str_to_int, 'str_to_int')

    assert check.dict_param({}, 'str_to_int', key_type=str, value_type=int) == {}
    assert check.dict_param({}, 'str_to_int', value_type=int) == {}
    assert check.dict_param({}, 'str_to_int', key_type=str) == {}
    assert check.dict_param({}, 'str_to_int') == {}

    class Wrong:
        pass

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, 'str_to_int', key_type=Wrong, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, 'str_to_int', key_type=Wrong, value_type=int)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, 'str_to_int', key_type=str, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, 'str_to_int', key_type=Wrong)

    with pytest.raises(CheckError):
        assert check.dict_param(str_to_int, 'str_to_int', value_type=Wrong)


def test_opt_dict_param_with_type():
    str_to_int = {'str': 1}
    assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=str, value_type=int)
    assert check.opt_dict_param(str_to_int, 'str_to_int', value_type=int)
    assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=str)
    assert check.opt_dict_param(str_to_int, 'str_to_int')

    assert check.opt_dict_param({}, 'str_to_int', key_type=str, value_type=int) == {}
    assert check.opt_dict_param({}, 'str_to_int', value_type=int) == {}
    assert check.opt_dict_param({}, 'str_to_int', key_type=str) == {}
    assert check.opt_dict_param({}, 'str_to_int') == {}

    assert check.opt_dict_param(None, 'str_to_int', key_type=str, value_type=int) == {}
    assert check.opt_dict_param(None, 'str_to_int', value_type=int) == {}
    assert check.opt_dict_param(None, 'str_to_int', key_type=str) == {}
    assert check.opt_dict_param(None, 'str_to_int') == {}

    class Wrong:
        pass

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=Wrong, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=Wrong, value_type=int)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=str, value_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, 'str_to_int', key_type=Wrong)

    with pytest.raises(CheckError):
        assert check.opt_dict_param(str_to_int, 'str_to_int', value_type=Wrong)


def test_opt_dict_param():
    assert check.opt_dict_param(None, 'opt_dict_param') == {}
    assert check.opt_dict_param({}, 'opt_dict_param') == {}
    ddict = {'a': 2}
    assert check.opt_dict_param(ddict, 'opt_dict_param') == ddict

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(0, 'opt_dict_param')

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(1, 'opt_dict_param')

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param('foo', 'opt_dict_param')

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param(['foo'], 'opt_dict_param')

    with pytest.raises(ParameterCheckError):
        check.opt_dict_param([], 'opt_dict_param')


def test_str_param():
    assert check.str_param('a', 'str_param') == 'a'
    assert check.str_param('', 'str_param') == ''
    assert check.str_param(u'a', 'unicode_param') == u'a'

    with pytest.raises(ParameterCheckError):
        check.str_param(None, 'str_param')

    with pytest.raises(ParameterCheckError):
        check.str_param(0, 'str_param')

    with pytest.raises(ParameterCheckError):
        check.str_param(1, 'str_param')


def test_opt_str_param():
    assert check.opt_str_param('a', 'str_param') == 'a'
    assert check.opt_str_param('', 'str_param') == ''
    assert check.opt_str_param(u'a', 'unicode_param') == u'a'
    assert check.opt_str_param(None, 'str_param') is None
    assert check.opt_str_param(None, 'str_param', 'foo') == 'foo'

    with pytest.raises(ParameterCheckError):
        check.opt_str_param(0, 'str_param')

    with pytest.raises(ParameterCheckError):
        check.opt_str_param(1, 'str_param')


def test_bool_param():
    assert check.bool_param(True, 'b') is True
    assert check.bool_param(False, 'b') is False

    with pytest.raises(ParameterCheckError):
        check.bool_param(None, 'b')

    with pytest.raises(ParameterCheckError):
        check.bool_param(0, 'b')

    with pytest.raises(ParameterCheckError):
        check.bool_param('val', 'b')


def test_opt_bool_param():
    assert check.opt_bool_param(True, 'b') is True
    assert check.opt_bool_param(False, 'b') is False
    assert check.opt_bool_param(None, 'b') is None
    assert check.opt_bool_param(None, 'b', True) is True
    assert check.opt_bool_param(None, 'b', False) is False

    with pytest.raises(ParameterCheckError):
        check.opt_bool_param(0, 'b')

    with pytest.raises(ParameterCheckError):
        check.opt_bool_param('val', 'b')


def test_callable_param():
    lamb = lambda: 1
    assert check.callable_param(lamb, 'lamb') == lamb

    with pytest.raises(ParameterCheckError):
        check.callable_param(None, 'lamb')

    with pytest.raises(ParameterCheckError):
        check.callable_param(2, 'lamb')


def test_opt_callable_param():
    lamb = lambda: 1
    assert check.opt_callable_param(lamb, 'lamb') == lamb
    assert check.opt_callable_param(None, 'lamb') is None
    assert check.opt_callable_param(None, 'lamb', default=None) is None
    assert check.opt_callable_param(None, 'lamb', default=lamb) == lamb

    with pytest.raises(ParameterCheckError):
        check.opt_callable_param(2, 'lamb')


def test_param_invariant():
    check.param_invariant(True, 'some_param')
    num_to_check = 1
    check.param_invariant(num_to_check == 1, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant(num_to_check == 2, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant(False, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant(0, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant(1, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant('', 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant('1kjkjsf', 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant({}, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant({234: '1kjkjsf'}, 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant([], 'some_param')

    with pytest.raises(ParameterCheckError):
        check.param_invariant([234], 'some_param')


def test_string_elem():
    ddict = {'a_str': 'a', 'a_num': 1, 'a_none': None}

    assert check.str_elem(ddict, 'a_str') == 'a'

    with pytest.raises(ElementCheckError):
        assert check.str_elem(ddict, 'a_none')

    with pytest.raises(ElementCheckError):
        check.str_elem(ddict, 'a_num')


def test_bool_elem():
    ddict = {'a_true': True, 'a_str': 'a', 'a_num': 1, 'a_none': None}

    assert check.bool_elem(ddict, 'a_true') is True

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, 'a_none')

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, 'a_num')

    with pytest.raises(ElementCheckError):
        check.bool_elem(ddict, 'a_str')


def test_invariant():
    assert check.invariant(True)

    with pytest.raises(CheckError):
        check.invariant(False)

    with pytest.raises(CheckError, match='Some Unique String'):
        check.invariant(False, 'Some Unique String')

    empty_list = []

    with pytest.raises(CheckError, match='Invariant condition must be boolean'):
        check.invariant(empty_list)


def test_failed():
    with pytest.raises(CheckError, match='some desc'):
        check.failed('some desc')

    with pytest.raises(CheckError, match='must be a string'):
        check.failed(0)


def test_not_implemented():
    with pytest.raises(NotImplementedCheckError, match='some string'):
        check.not_implemented('some string')


def test_inst():
    class Foo:
        pass

    class Bar:
        pass

    obj = Foo()

    assert check.inst(obj, Foo) == obj

    with pytest.raises(CheckError, match='not a Bar'):
        check.inst(Foo(), Bar)


def test_inst_param():
    class Foo:
        pass

    class Bar:
        pass

    obj = Foo()

    assert check.inst_param(obj, 'obj', Foo) == obj

    with pytest.raises(ParameterCheckError, match='not a Bar'):
        check.inst_param(None, 'obj', Bar)

    with pytest.raises(ParameterCheckError, match='not a Bar'):
        check.inst_param(Foo(), 'obj', Bar)


def test_opt_inst_param():
    class Foo:
        pass

    class Bar:
        pass

    obj = Foo()

    assert check.opt_inst_param(obj, 'obj', Foo) == obj
    assert check.opt_inst_param(None, 'obj', Foo) is None
    assert check.opt_inst_param(None, 'obj', Bar) is None

    with pytest.raises(ParameterCheckError, match='not a Bar'):
        check.opt_inst_param(Foo(), 'obj', Bar)

    # check defaults

    default_obj = Foo()

    assert check.opt_inst_param(None, 'obj', Foo, default_obj) is default_obj


def test_dict_elem():
    dict_value = {'blah': 'blahblah'}
    ddict = {'dictkey': dict_value, 'stringkey': 'A', 'nonekey': None}

    assert check.dict_elem(ddict, 'dictkey') == dict_value

    with pytest.raises(CheckError):
        check.dict_elem(ddict, 'stringkey')

    with pytest.raises(CheckError):
        check.dict_elem(ddict, 'nonekey')

    with pytest.raises(CheckError):
        check.dict_elem(ddict, 'nonexistantkey')


def test_opt_dict_elem():
    dict_value = {'blah': 'blahblah'}
    ddict = {'dictkey': dict_value, 'stringkey': 'A', 'nonekey': None}

    assert check.opt_dict_elem(ddict, 'dictkey') == dict_value
    assert check.opt_dict_elem(ddict, 'nonekey') == {}
    assert check.opt_dict_elem(ddict, 'nonexistantkey') == {}

    with pytest.raises(CheckError):
        check.opt_dict_elem(ddict, 'stringkey')


def test_opt_list_elem():
    list_value = ['blah', 'blahblah']
    ddict = {'listkey': list_value, 'stringkey': 'A', 'nonekey': None}

    assert check.opt_list_elem(ddict, 'listkey') == list_value
    assert check.opt_list_elem(ddict, 'nonekey') == []
    assert check.opt_list_elem(ddict, 'nonexistantkey') == []

    with pytest.raises(CheckError):
        check.opt_list_elem(ddict, 'stringkey')


def test_not_none_param():
    assert check.not_none_param(1, 'fine')
    check.not_none_param(0, 'zero is fine')
    check.not_none_param('', 'empty str is fine')

    with pytest.raises(CheckError):
        check.not_none_param(None, 'none fails')


def test_is_callable():
    def fn():
        pass

    assert check.is_callable(fn) == fn
    assert check.is_callable(lambda: None)

    with pytest.raises(CheckError):
        check.is_callable(None)

    with pytest.raises(CheckError):
        check.is_callable(1)


def test_tuple_param():
    assert check.tuple_param((1, 2), 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param(None, 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param(1, 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param([1], 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param({1: 2}, 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param('kdjfkd', 'something')


def test_opt_tuple_param():
    assert check.opt_tuple_param((1, 2), 'something')
    assert check.opt_tuple_param(None, 'something') is None
    assert check.opt_tuple_param(None, 'something', (2)) == (2)

    with pytest.raises(CheckError):
        assert check.tuple_param(1, 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param([1], 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param({1: 2}, 'something')

    with pytest.raises(CheckError):
        assert check.tuple_param('kdjfkd', 'something')


def test_type_param():
    class Bar:
        pass

    assert check.type_param(int, 'foo')
    assert check.type_param(Bar, 'foo')

    with pytest.raises(CheckError):
        check.type_param(None, 'foo')

    with pytest.raises(CheckError):
        check.type_param(check, 'foo')

    with pytest.raises(CheckError):
        check.type_param(234, 'foo')

    with pytest.raises(CheckError):
        check.type_param('bar', 'foo')

    with pytest.raises(CheckError):
        check.type_param(Bar(), 'foo')



import sqlalchemy as sa
import dagster
from dagster import (config, InputDefinition)
import dagster.sqlalchemy_kernel as dagster_sa
from dagster.utils.test import script_relative_path

from dagster.sqlalchemy_kernel.subquery_builder_experimental import sql_file_solid


def in_mem_engine():
    engine = sa.create_engine('sqlite://', echo=False)
    return engine


def in_mem_context():
    return dagster_sa.create_sql_alchemy_context_from_engine(engine=in_mem_engine())


def pipeline_engine(pipeline_result):
    return pipeline_result.context.resources.sa.engine


def create_persisted_context():
    full_path = script_relative_path('testdb.db')
    engine = sa.create_engine(f'sqlite:///{full_path}', echo=False)
    return dagster_sa.create_sql_alchemy_context_from_engine(engine=engine)


def create_mem_sql_pipeline_context_tuple(solids):
    default_def = dagster.PipelineContextDefinition(
        argument_def_dict={},
        context_fn=lambda _pipeline, _args: in_mem_context(),
    )
    persisted_def = dagster.PipelineContextDefinition(
        argument_def_dict={},
        context_fn=lambda _pipeline, _args: create_persisted_context(),
    )
    return dagster.PipelineDefinition(
        solids=solids, context_definitions={
            'default': default_def,
            'persisted': persisted_def
        }
    )


def _get_sql_script_path(name):
    return script_relative_path(f'../../sql_project_example/sql_files/{name}.sql')


def _get_project_solid(name, inputs=None):
    return sql_file_solid(_get_sql_script_path(name), inputs=inputs)


def test_sql_create_tables():
    create_all_tables_solids = _get_project_solid('create_all_tables')

    pipeline = create_mem_sql_pipeline_context_tuple(solids=[create_all_tables_solids])

    pipeline_result = dagster.execute_pipeline(pipeline, environment=config.Environment.empty())
    assert pipeline_result.success

    assert set(pipeline_engine(pipeline_result).table_names()) == set(
        ['num_table', 'sum_sq_table', 'sum_table']
    )


def test_sql_populate_tables():
    create_all_tables_solids = _get_project_solid('create_all_tables')

    populate_num_table_solid = _get_project_solid(
        'populate_num_table',
        inputs=[
            InputDefinition(create_all_tables_solids.name, depends_on=create_all_tables_solids)
        ]
    )

    pipeline = create_mem_sql_pipeline_context_tuple(
        solids=[create_all_tables_solids, populate_num_table_solid]
    )

    pipeline_result = dagster.execute_pipeline(pipeline, environment=config.Environment.empty())

    assert pipeline_result.success

    assert pipeline_engine(pipeline_result).execute('SELECT * FROM num_table').fetchall() == [
        (1, 2), (3, 4)
    ]


def create_full_pipeline():
    create_all_tables_solids = _get_project_solid('create_all_tables')

    populate_num_table_solid = _get_project_solid(
        'populate_num_table',
        inputs=[InputDefinition('create_all_tables_solids', depends_on=create_all_tables_solids)],
    )

    insert_into_sum_table_solid = _get_project_solid(
        'insert_into_sum_table',
        inputs=[InputDefinition('populate_num_table_solid', depends_on=populate_num_table_solid)],
    )

    insert_into_sum_sq_table_solid = _get_project_solid(
        'insert_into_sum_sq_table',
        inputs=[
            InputDefinition('insert_into_sum_sq_table', depends_on=insert_into_sum_table_solid)
        ],
    )

    return create_mem_sql_pipeline_context_tuple(
        solids=[
            create_all_tables_solids,
            populate_num_table_solid,
            insert_into_sum_table_solid,
            insert_into_sum_sq_table_solid,
        ],
    )


def test_full_in_memory_pipeline():

    pipeline = create_full_pipeline()
    pipeline_result = dagster.execute_pipeline(pipeline, environment=config.Environment.empty())
    assert pipeline_result.success

    engine = pipeline_engine(pipeline_result)
    assert engine.execute('SELECT * FROM num_table').fetchall() == [(1, 2), (3, 4)]
    assert engine.execute('SELECT * FROM sum_table').fetchall() == [(1, 2, 3), (3, 4, 7)]
    assert engine.execute('SELECT * FROM sum_sq_table').fetchall() == [(1, 2, 3, 9), (3, 4, 7, 49)]


def test_full_persisted_pipeline():
    pipeline = create_full_pipeline()
    pipeline_result = dagster.execute_pipeline(
        pipeline,
        environment=config.Environment(
            sources={}, context=config.Context(name='persisted', args={})
        )
    )

    assert pipeline_result.success



import pytest

import dagster

from dagster.core.definitions import (
    ExpectationDefinition,
    ExpectationResult,
    ArgumentDefinition,
)
from dagster.core.execution import (
    _execute_input_expectation, DagsterUserCodeExecutionError, ExecutionContext
)
from dagster.core import types

from dagster.utils.compatability import create_custom_source_input


def create_test_context():
    return ExecutionContext()


def test_basic_failing_input_expectation():
    def failing_expectation(_context, _info, _some_input):
        return ExpectationResult(success=False, message='Some failure')

    some_input = create_custom_source_input(
        name='some_input',
        source_fn=lambda arg_dict: [{'key': arg_dict['str_arg']}],
        argument_def_dict={'str_arg': ArgumentDefinition(types.String)},
        expectations=[
            ExpectationDefinition(name='failing', expectation_fn=failing_expectation)
        ]
    )

    expectation_def = some_input.expectations[0]

    info = create_dummy_info(expectation_def)

    result = _execute_input_expectation(create_test_context(), info, 'some_value')

    assert isinstance(result, ExpectationResult)
    assert not result.success
    assert result.message == 'Some failure'


def create_dummy_info(expectation_def):
    info = dagster.InputExpectationInfo(
        solid=dagster.SolidDefinition(
            name='dummy',
            inputs=[],
            transform_fn=lambda _context, _args: None,
            output=dagster.OutputDefinition(),
        ),
        input_def=dagster.InputDefinition(name='dummy_input'),
        expectation_def=expectation_def,
    )
    return info


def test_basic_passing_input_expectation():
    def passing_expectation(_context, _info, _some_input):
        return ExpectationResult(success=True, message='yayayaya')

    some_input = create_custom_source_input(
        name='some_input',
        source_fn=lambda arg_dict: [{'key': arg_dict['str_arg']}],
        argument_def_dict={'str_arg': ArgumentDefinition(types.String)},
        expectations=[
            ExpectationDefinition(name='passing', expectation_fn=passing_expectation)
        ]
    )

    result = _execute_input_expectation(
        create_test_context(), create_dummy_info(some_input.expectations[0]), 'some_value'
    )

    assert isinstance(result, ExpectationResult)
    assert result.success
    assert result.message == 'yayayaya'


def test_input_expectation_user_error():
    def throwing(_context, _info, _something):
        raise Exception('nope')

    failing_during_expectation_input = create_custom_source_input(
        name='failing_during_expectation',
        source_fn=lambda arg_dict: [{'key': arg_dict['str_arg']}],
        argument_def_dict={'str_arg': ArgumentDefinition(types.String)},
        expectations=[
            ExpectationDefinition(name='passing', expectation_fn=throwing)
        ]
    )

    with pytest.raises(DagsterUserCodeExecutionError):
        _execute_input_expectation(
            create_test_context(),
            create_dummy_info(failing_during_expectation_input.expectations[0]),
            'some_value',
        )



from future.utils import raise_with_traceback
from six import string_types


class CheckError(Exception):
    pass


class ParameterCheckError(CheckError):
    pass


class ElementCheckError(CheckError):
    pass


class NotImplementedCheckError(CheckError):
    pass


def _param_type_mismatch_exception(obj, ttype, param_name):
    return ParameterCheckError(
        'Param "{name}" is not a {type}. Got {obj} with is type {obj_type}.'.format(
            name=param_name, obj=repr(obj), type=ttype.__name__, obj_type=type(obj).__name__
        )
    )


def _type_mismatch_error(obj, ttype, desc):
    if desc:
        return CheckError(
            'Object {obj} is not a {type}. Got {obj} with type {obj_type}. Desc: {desc}'.format(
                obj=repr(obj), type=ttype.__name__, obj_type=type(obj).__name__, desc=desc
            )
        )
    else:
        return CheckError(
            'Object {obj} is not a {type}. Got {obj} with type {obj_type}.'.format(
                obj=repr(obj), type=ttype.__name__, obj_type=type(obj).__name__
            )
        )


def _not_callable_exception(obj, param_name):
    return ParameterCheckError(
        'Param "{name}" is not callable. Got {obj} with type {obj_type}.'.format(
            name=param_name, obj=repr(obj), obj_type=type(obj)
        )
    )


def _param_invariant_exception(param_name, desc):
    return ParameterCheckError(
        'Invariant violation for parameter {param_name}. Description: {desc}'.format(
            param_name=param_name, desc=desc
        )
    )


def failed(desc):
    if not _is_str(desc):
        raise_with_traceback(CheckError('desc argument must be a string'))

    raise_with_traceback(CheckError('Failure condition: {desc}'.format(desc=desc)))


def not_implemented(desc):
    if not _is_str(desc):
        raise_with_traceback(CheckError('desc argument must be a string'))

    raise_with_traceback(NotImplementedCheckError('Not implemented: {desc}'.format(desc=desc)))


def inst(obj, ttype, desc=None):
    if not isinstance(obj, ttype):
        raise_with_traceback(_type_mismatch_error(obj, ttype, desc))
    return obj


def is_callable(obj):
    if not callable(obj):
        raise_with_traceback(CheckError('must be callable'))

    return obj


def not_none_param(obj, param_name):
    if obj is None:
        raise_with_traceback(
            _param_invariant_exception(
                param_name, 'Param {param_name} cannot be none'.format(param_name=param_name)
            )
        )
    return obj


def invariant(condition, desc=None):
    if not isinstance(condition, bool):
        raise_with_traceback(CheckError('Invariant condition must be boolean'))

    if not condition:
        if desc:
            raise_with_traceback(
                CheckError('Invariant failed. Description: {desc}'.format(desc=desc))
            )
        else:
            raise_with_traceback(CheckError('Invariant failed.'))

    return True


def param_invariant(condition, param_name, desc=None):
    if not isinstance(condition, bool):
        raise_with_traceback(ParameterCheckError('Invariant condition must be boolean'))

    if not condition:
        raise_with_traceback(_param_invariant_exception(param_name, desc))


def inst_param(obj, param_name, ttype):
    if not isinstance(obj, ttype):
        raise_with_traceback(_param_type_mismatch_exception(obj, ttype, param_name))
    return obj


def opt_inst_param(obj, param_name, ttype, default=None):
    if obj is not None and not isinstance(obj, ttype):
        raise_with_traceback(_param_type_mismatch_exception(obj, ttype, param_name))
    return default if obj is None else obj


def callable_param(obj, param_name):
    if not callable(obj):
        raise_with_traceback(_not_callable_exception(obj, param_name))
    return obj


def opt_callable_param(obj, param_name, default=None):
    if obj is not None and not callable(obj):
        raise_with_traceback(_not_callable_exception(obj, param_name))
    return default if obj is None else obj


def int_param(obj, param_name):
    if not isinstance(obj, int):
        raise_with_traceback(_param_type_mismatch_exception(obj, int, param_name))
    return obj


def opt_int_param(obj, param_name):
    if obj is not None and not isinstance(obj, int):
        raise_with_traceback(_param_type_mismatch_exception(obj, int, param_name))
    return obj


def _is_str(obj):
    return isinstance(obj, string_types)


def str_param(obj, param_name):
    if not _is_str(obj):
        raise_with_traceback(_param_type_mismatch_exception(obj, str, param_name))
    return obj


def opt_str_param(obj, param_name, default=None):
    if obj is not None and not isinstance(obj, string_types):
        raise_with_traceback(_param_type_mismatch_exception(obj, str, param_name))
    return default if obj is None else obj


def bool_param(obj, param_name):
    if not isinstance(obj, bool):
        raise_with_traceback(_param_type_mismatch_exception(obj, bool, param_name))
    return obj


def opt_bool_param(obj, param_name, default=None):
    if obj is not None and not isinstance(obj, bool):
        raise_with_traceback(_param_type_mismatch_exception(obj, bool, param_name))
    return default if obj is None else obj


def list_param(obj_list, param_name, of_type=None):
    if not isinstance(obj_list, list):
        raise_with_traceback(_param_type_mismatch_exception(obj_list, list, param_name))
    if of_type:
        for obj in obj_list:
            if not isinstance(obj, of_type):
                raise_with_traceback(CheckError(
                    'Member of list mismatches type. Got {obj_repr}'.format(obj_repr=repr(obj))
                ))
    return obj_list


def tuple_param(obj, param_name):
    if not isinstance(obj, tuple):
        raise_with_traceback(_param_type_mismatch_exception(obj, tuple, param_name))
    return obj


def opt_tuple_param(obj, param_name, default=None):
    if obj is not None and not isinstance(obj, tuple):
        raise_with_traceback(_param_type_mismatch_exception(obj, tuple, param_name))
    return default if obj is None else obj


def opt_list_param(obj_list, param_name, of_type=None):
    if obj_list is not None and not isinstance(obj_list, list):
        raise_with_traceback(_param_type_mismatch_exception(obj_list, list, param_name))
    if not obj_list:
        return []
    if of_type:
        for obj in obj_list:
            if not isinstance(obj, of_type):
                raise_with_traceback(CheckError(
                    'Member of list mismatches type. Got {obj_repr}'.format(obj_repr=repr(obj))
                ))
    return obj_list


def dict_param(obj, param_name, key_type=None, value_type=None):
    if not isinstance(obj, dict):
        raise_with_traceback(_param_type_mismatch_exception(obj, dict, param_name))

    if not (key_type or value_type):
        return obj

    return _check_key_value_types(obj, key_type, value_type)


def type_param(obj, param_name):
    if not isinstance(obj, type):
        raise_with_traceback(_param_type_mismatch_exception(obj, type, param_name))
    return obj


def _check_key_value_types(obj, key_type, value_type):
    for key, value in obj.items():
        if key_type and not isinstance(key, key_type):
            raise_with_traceback(CheckError('Key in dictionary mismatches type'))
        if value_type and not isinstance(value, value_type):
            raise_with_traceback(CheckError('Value in dictionary mismatches type'))
    return obj


def opt_dict_param(obj, param_name, key_type=None, value_type=None):
    if obj is not None and not isinstance(obj, dict):
        raise_with_traceback(_param_type_mismatch_exception(obj, dict, param_name))

    if not obj:
        return {}

    return _check_key_value_types(obj, key_type, value_type)


def _element_check_error(key, value, ddict, ttype):
    return ElementCheckError(
        'Value {value} from key {key} is not a {ttype}. Dict: {ddict}'.format(
            key=key, value=repr(value), ddict=repr(ddict), ttype=repr(ttype)
        )
    )


def list_elem(ddict, key):
    dict_param(ddict, 'ddict')
    str_param(key, 'key')

    value = ddict[key]
    if not isinstance(value, list):
        raise_with_traceback(_element_check_error(key, value, ddict, list))
    return value


def opt_list_elem(ddict, key):
    dict_param(ddict, 'ddict')
    str_param(key, 'key')

    value = ddict.get(key)

    if value is None:
        return []

    if not isinstance(value, list):
        raise_with_traceback(_element_check_error(key, value, ddict, list))
    return value


def dict_elem(ddict, key):
    dict_param(ddict, 'ddict')
    str_param(key, 'key')

    if key not in ddict:
        raise_with_traceback(
            CheckError('{key} not present in dictionary {ddict}'.format(key=key, ddict=ddict))
        )

    value = ddict[key]
    if not isinstance(value, dict):
        raise_with_traceback(_element_check_error(key, value, ddict, dict))
    return value


def opt_dict_elem(ddict, key):
    dict_param(ddict, 'ddict')
    str_param(key, 'key')

    value = ddict.get(key)

    if value is None:
        return {}

    if not isinstance(value, dict):
        raise_with_traceback(_element_check_error(key, value, ddict, list))

    return value


def bool_elem(ddict, key):
    dict_param(ddict, 'ddict')
    str_param(key, 'key')

    value = ddict[key]
    if not isinstance(value, bool):
        raise_with_traceback(_element_check_error(key, value, ddict, bool))
    return value


def opt_str_elem(ddict, key):
    dict_param(ddict, 'ddict')
    str_param(key, 'key')

    value = ddict.get(key)
    if value is None:
        return None
    if not _is_str(value):
        raise_with_traceback(_element_check_error(key, value, ddict, str))
    return value


def str_elem(ddict, key):
    dict_param(ddict, 'ddict')
    str_param(key, 'key')

    value = ddict[key]
    if not _is_str(value):
        raise_with_traceback(_element_check_error(key, value, ddict, str))
    return value



import pytest

import dagster
from dagster.core.definitions import (
    ExpectationDefinition,
    ExpectationResult,
)
from dagster.core.execution import (
    _execute_output_expectation, DagsterUserCodeExecutionError, ExecutionContext
)


def create_test_context():
    return ExecutionContext()


def create_dummy_output_info(expect_def):
    return dagster.OutputExpectationInfo(
        solid=dagster.SolidDefinition(
            name='dummy',
            inputs=[],
            transform_fn=lambda _context, _args: None,
            output=dagster.OutputDefinition(expectations=[expect_def])
        ),
        expectation_def=expect_def
    )


def test_basic_failing_output_expectation():
    def failing(_context, _info, _output):
        return ExpectationResult(
            success=False,
            message='some message',
        )

    result = _execute_output_expectation(
        create_test_context(), create_dummy_output_info(ExpectationDefinition('failing', failing)),
        'not used'
    )

    assert isinstance(result, ExpectationResult)
    assert not result.success
    assert result.message == 'some message'


def test_basic_passing_output_expectation():
    def success(_context, _info, _output):
        return ExpectationResult(
            success=True,
            message='yay',
        )

    expectation = ExpectationDefinition('success', success)
    result = _execute_output_expectation(
        create_test_context(),
        create_dummy_output_info(expectation),
        'not used',
    )

    assert isinstance(result, ExpectationResult)
    assert result.success
    assert result.message == 'yay'


def test_output_expectation_user_error():
    def throwing(_output):
        raise Exception('user error')

    with pytest.raises(DagsterUserCodeExecutionError):
        _execute_output_expectation(
            create_test_context(),
            create_dummy_output_info(ExpectationDefinition('throwing', throwing)), 'not used'
        )



import pytest

from dagster.core.definitions import check_valid_name
from dagster.core.errors import DagsterInvalidDefinitionError


def test_check_valid_name():
    assert check_valid_name('a') == 'a'

    with pytest.raises(DagsterInvalidDefinitionError):
        assert check_valid_name('has a space')

    with pytest.raises(DagsterInvalidDefinitionError):
        assert check_valid_name('')

    with pytest.raises(DagsterInvalidDefinitionError):
        assert check_valid_name('context')



from enum import Enum

from dagster import check


class DagsterExecutionFailureReason(Enum):
    USER_CODE_ERROR = 'USER_CODE_ERROR'
    FRAMEWORK_ERROR = 'FRAMEWORK_ERROR'
    EXPECTATION_FAILURE = 'EXPECATION_FAILURE'


class DagsterError(Exception):
    pass


class DagsterUserError(DagsterError):
    pass


class DagsterInvalidDefinitionError(DagsterUserError):
    '''Indicates that some violation of the definition rules has been violated by the user'''
    pass


class DagsterInvariantViolationError(DagsterUserError):
    '''Indicates the user has violated a well-defined invariant that can only be deteremined
    at runtime.
    '''
    pass


class DagsterTypeError(DagsterUserError):
    '''Indicates an error in the solid type system (e.g. mismatched arguments)'''
    pass


class DagsterUserCodeExecutionError(DagsterUserError):
    '''Indicates that user space code has raised an error'''

    def __init__(self, *args, user_exception, original_exc_info, **kwargs):
        # original_exc_info should be gotten from a sys.exc_info() call at the
        # callsite inside of the exception handler. this will allow consuming
        # code to *re-raise* the user error in it's original format
        # for cleaner error reporting that does not have framework code in it
        super().__init__(*args, **kwargs)

        self.user_exception = check.opt_inst_param(user_exception, 'user_exception', Exception)
        self.original_exc_info = original_exc_info


class DagsterExpectationFailedError(DagsterError):
    '''Thrown with pipeline configured to throw on expectation failure'''

    def __init__(self, execution_result, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # FIXME: need to reorganize to fix this circular dep
        # Probable fix is to move all "execution result" objects
        # to definitions
        import dagster.core.execution
        self.execution_result = check.inst_param(
            execution_result,
            'execution_result',
            dagster.core.execution.SolidExecutionResult,
        )



import pytest

from dagster.core import types

from dagster.core.definitions import (
    SolidDefinition,
    OutputDefinition,
    MaterializationDefinition,
    SourceDefinition,
    ArgumentDefinition,
)

from dagster.core.execution import (
    _execute_core_transform,
    DagsterTypeError,
    ExecutionContext,
    _execute_materialization,
    _read_source,
)

from dagster.utils.compatability import create_custom_source_input


def create_test_context():
    return ExecutionContext()


def _read_new_single_source_input(context, new_input, arg_dict):
    assert len(new_input.sources) == 1
    return _read_source(context, new_input.sources[0], arg_dict)


def test_read_source():
    expected_output = [{'data_key': 'data_value'}]
    some_input = create_custom_source_input(
        name='some_input',
        source_fn=lambda context, arg_dict: expected_output,
        argument_def_dict={}
    )

    output = _read_source(create_test_context(), some_input.sources[0], {})

    assert output == expected_output


def test_source_arg_mismiatch():
    extra_arg_source = create_custom_source_input(
        name='some_input', source_fn=lambda context, arg_dict: [], argument_def_dict={}
    ).sources[0]

    with pytest.raises(DagsterTypeError):
        _read_source(create_test_context(), extra_arg_source, {'extra_arg': None})

    some_input_with_arg = create_custom_source_input(
        name='some_input_with_arg',
        source_fn=lambda context, arg_dict: [],
        argument_def_dict={'in_arg': ArgumentDefinition(types.String)},
    )

    with pytest.raises(DagsterTypeError):
        _read_new_single_source_input(create_test_context(), some_input_with_arg, {})


def test_source_arg_type_mismatch():
    some_input_with_arg = create_custom_source_input(
        name='some_input_with_arg',
        source_fn=lambda context, arg_dict: [],
        argument_def_dict={'in_arg': ArgumentDefinition(types.String)},
    )

    with pytest.raises(DagsterTypeError):
        _read_new_single_source_input(create_test_context(), some_input_with_arg, {'in_arg': 1})


def test_source_int_type():
    int_arg_source = SourceDefinition(
        source_type='SOMETHING',
        source_fn=lambda _context, _args: True,
        argument_def_dict={'an_int': ArgumentDefinition(types.Int)},
    )
    assert _read_source(create_test_context(), int_arg_source, {'an_int': 0})
    assert _read_source(create_test_context(), int_arg_source, {'an_int': 1})
    assert _read_source(create_test_context(), int_arg_source, {'an_int': None})

    with pytest.raises(DagsterTypeError):
        _read_source(create_test_context(), int_arg_source, {'an_int': 'not_an_int'})


def test_source_bool_type():
    bool_arg_source = SourceDefinition(
        source_type='SOMETHING',
        source_fn=lambda _context, _args: True,
        argument_def_dict={'an_bool': ArgumentDefinition(types.Bool)},
    )

    assert _read_source(create_test_context(), bool_arg_source, {'an_bool': True})
    assert _read_source(create_test_context(), bool_arg_source, {'an_bool': False})
    assert _read_source(create_test_context(), bool_arg_source, {'an_bool': None})

    with pytest.raises(DagsterTypeError):
        _read_source(create_test_context(), bool_arg_source, {'an_bool': 'not_an_bool'})

    with pytest.raises(DagsterTypeError):
        _read_source(create_test_context(), bool_arg_source, {'an_bool': 0})


def noop_output():
    return OutputDefinition()


def test_materialize_output():
    some_input = create_custom_source_input(
        name='some_input',
        source_fn=lambda context, arg_dict: [{'data_key': 'data_value'}],
        argument_def_dict={},
    )

    def tranform_fn_inst(_context, args):
        args['some_input'][0]['data_key'] = 'new_value'
        return args['some_input']

    single_solid = SolidDefinition(
        name='some_node',
        inputs=[some_input],
        transform_fn=tranform_fn_inst,
        output=noop_output(),
    )

    value = _read_new_single_source_input(create_test_context(), some_input, {})

    output = _execute_core_transform(
        create_test_context(),
        single_solid.transform_fn,
        {'some_input': value},
    )

    assert output == [{'data_key': 'new_value'}]


def test_materialize_input_with_args():
    some_input = create_custom_source_input(
        name='some_input',
        source_fn=lambda context, arg_dict: [{'key': arg_dict['str_arg']}],
        argument_def_dict={'str_arg': ArgumentDefinition(types.String)},
    )

    output = _read_new_single_source_input(
        create_test_context(), some_input, {'str_arg': 'passed_value'}
    )
    expected_output = [{'key': 'passed_value'}]
    assert output == expected_output


def test_execute_output_with_args():
    test_output = {}

    def materialization_fn_inst(context, arg_dict, value):
        assert isinstance(context, ExecutionContext)
        assert isinstance(arg_dict, dict)
        test_output['thedata'] = value
        test_output['thearg'] = arg_dict['out_arg']

    materialization = MaterializationDefinition(
        name='CUSTOM',
        materialization_fn=materialization_fn_inst,
        argument_def_dict={'out_arg': ArgumentDefinition(types.String)}
    )

    _execute_materialization(
        create_test_context(), materialization, {'out_arg': 'the_out_arg'}, [{
            'key': 'value'
        }]
    )


def test_execute_materialization_arg_mismatch():
    materialization = MaterializationDefinition(
        name='CUSTOM',
        materialization_fn=lambda out, dict: [],
        argument_def_dict={'out_arg': ArgumentDefinition(types.String)}
    )

    with pytest.raises(DagsterTypeError):
        _execute_materialization(create_test_context(), materialization, arg_dict={}, value=[{}])

    with pytest.raises(DagsterTypeError):
        _execute_materialization(
            create_test_context(),
            materialization,
            arg_dict={
                'out_arg': 'foo',
                'extra_arg': 'bar'
            },
            value=[{}]
        )


def test_execute_materialization_arg_type_mismatch():
    custom_output = MaterializationDefinition(
        name='CUSTOM',
        materialization_fn=lambda out, dict: [],
        argument_def_dict={'out_arg': ArgumentDefinition(types.String)}
    )

    with pytest.raises(DagsterTypeError):
        _execute_materialization(
            create_test_context(), custom_output, arg_dict={'out_arg': 1}, value=[{}]
        )
